#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2016 Simon McVittie <smcv@debian.org>
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import sys
import unittest

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))

if 'GDP_UNINSTALLED' not in os.environ:
    sys.path.insert(0, '/usr/share/game-data-packager')
    sys.path.insert(0, '/usr/share/games/game-data-packager')

from game_data_packager.data import (PackageRelation)  # noqa: E402
from game_data_packager.packaging.deb import (DebPackaging)  # noqa: E402


class DebTestCase(unittest.TestCase):
    def setUp(self) -> None:
        pass

    def test_multiarch(self) -> None:
        archs = [
            ('amd64', 'x86_64-linux-gnu'),
            ('arm64', 'aarch64-linux-gnu'),
            ('armel', 'arm-linux-gnueabi'),
            ('armhf', 'arm-linux-gnueabihf'),
            ('i386', 'i386-linux-gnu'),
            ('mips64el', 'mips64el-linux-gnuabi64'),
            ('ppc64el', 'powerpc64le-linux-gnu'),
            ('riscv64', 'riscv64-linux-gnu'),
            ('s390x', 's390x-linux-gnu'),
            ('alpha', 'alpha-linux-gnu'),
            ('hppa', 'hppa-linux-gnu'),
            ('hurd-amd64', 'x86_64-gnu'),
            ('hurd-i386', 'i386-gnu'),
            ('loong64', 'loongarch64-linux-gnu'),
            ('m68k', 'm68k-linux-gnu'),
            ('powerpc', 'powerpc-linux-gnu'),
            ('ppc64', 'powerpc64-linux-gnu'),
            ('sh4', 'sh4-linux-gnu'),
            ('sparc64', 'sparc64-linux-gnu'),
            ('x32', 'x86_64-linux-gnux32'),
        ]

        for role in 'BUILD', 'HOST', 'TARGET':
            if (
                f'DEB_{role}_ARCH' in os.environ
                and f'DEB_{role}_MULTIARCH' in os.environ
            ):
                archs.append((
                    os.environ[f'DEB_{role}_ARCH'],
                    os.environ[f'DEB_{role}_MULTIARCH'],
                ))

        for arch, expected in archs:
            dp = DebPackaging(arch)
            self.assertEqual(
                dp.get_libdir(arch),
                f'/usr/lib/{expected}',
                f'expected {arch} -> {expected}',
            )

    def test_rename_package(self) -> None:
        dp = DebPackaging()

        def t(in_: str, out: str) -> None:
            self.assertEqual(dp.rename_package(in_), out)

        # generic
        t('libc.so.6', 'libc6')
        t('libalut.so.0', 'libalut0')
        t('libXxf86dga.so.1', 'libxxf86dga1')
        t('libopenal.so.1', 'libopenal1')
        t('libstdc++.so.6', 'libstdc++6')
        t('libdbus-1.so.3', 'libdbus-1-3')

        # special cases
        t('libSDL-1.2.so.0', 'libsdl1.2debian')
        t('libgcc_s.so.1', 'libgcc1')
        t('libjpeg.so.62', 'libjpeg62-turbo | libjpeg62')

    def test_relation(self) -> None:
        dp = DebPackaging()

        def t(
            in_: list[str | dict[str, str]],
            out: list[str]
        ) -> None:
            self.assertEqual(
                    sorted(dp.format_relations(map(PackageRelation, in_))),
                    out)

        t(['libc.so.6'], ['libc6'])
        t(['libc.so.6 (>= 2.19)'], ['libc6 (>= 2.19)'])
        t(['libjpeg.so.62'], ['libjpeg62-turbo | libjpeg62'])
        t(['libopenal.so.1 | bundled-openal'], ['libopenal1 | bundled-openal'])
        t(['libc.so.6', 'libopenal.so.1'], ['libc6', 'libopenal1'])
        t([dict(deb='foo', rpm='bar')], ['foo'])
        t([dict(deb='foo', rpm='bar', generic='baz')], ['foo'])
        t([dict(rpm='bar', generic='baz')], ['baz'])
        t([dict(rpm='bar')], [])

    def tearDown(self) -> None:
        pass


if __name__ == '__main__':
    from gdp_test_common import main
    main()



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2016 Simon McVittie <smcv@debian.org>
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import subprocess
import sys
import unittest
from tempfile import (TemporaryDirectory)

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))


class DownloadSharewareTestCase(unittest.TestCase):
    '''
    Test some cherry picked games that:
    - are freely downloadable (either demo or full version)
    - test various codepaths:
      - alternatives
      - archive recursion (zip in zip)
      - lha
      - id-shr-extract
    - doom_common.py plugin
    - are not too big
    '''

    def setUp(self) -> None:
        if 'GDP_MIRROR' in os.environ:
            self.downloads = None
            if os.environ['GDP_MIRROR'] == 'none':
                self.skipTest('GDP_MIRROR is set to "none"')
        else:
            try:
                from gi.repository import GLib
            except ImportError:
                self.skipTest('GLib g-i bindings not available')

            self.downloads = GLib.get_user_special_dir(
                    GLib.UserDirectory.DIRECTORY_DOWNLOAD)
            if (self.downloads is None or
                    not os.path.isdir(self.downloads)):
                self.skipTest('XDG download directory "{}" not found'.format(
                    self.downloads))

            os.environ['GDP_MIRROR'] = 'file://' + self.downloads

    def _test_one(self, game: str, files: list[str]) -> None:
        if self.downloads is not None:
            for filename in files:
                if not os.path.exists(os.path.join(self.downloads, filename)):
                    self.skipTest('download {} into {}'.format(filename,
                                  self.downloads))

        with TemporaryDirectory(prefix='gdptest.') as tmp:
            if 'GDP_UNINSTALLED' in os.environ:
                argv = [
                    os.path.join(
                        os.environ.get('GDP_BUILDDIR', 'out'),
                        'run-gdp-uninstalled',
                    )
                ]
            else:
                argv = ['game-data-packager']

            argv = argv + [
                '-d', tmp,
                '--no-compress',
                game,
                '--no-search',
            ]
            subprocess.check_call(argv)
            if 'GDP_TEST_ALL_FORMATS' in os.environ:
                for f in 'arch deb rpm'.split():
                    subprocess.check_call(argv + ['--target-format', f])

    def test_heretic(self) -> None:
        self._test_one('heretic', 'htic_v12.zip'.split())

    def test_rott(self) -> None:
        self._test_one('rott', '1rott13.zip'.split())

    def test_spear(self) -> None:
        self._test_one('spear-of-destiny', 'sodemo.zip destiny.zip'.split())

    def test_wolf3d(self) -> None:
        self._test_one('wolf3d', '1wolf14.zip'.split())

    def tearDown(self) -> None:
        pass


if __name__ == '__main__':
    if 'DEB_BUILD_TIME_TESTS' in os.environ:
        print('# SKIP: not doing test that requires network access at '
              'build-time')
    else:
        from gdp_test_common import main
        main()



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2024 Sébastien Noel <sebastien@twolife.be>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
import os
from collections.abc import Collection
from types import TracebackType

from . import (SimpleUnpackable)
from ..util import (mkdir_p)

PKG_MAGIC = 0x4f504b47

logger = logging.getLogger(__name__)


class D3Entry():
    """Data class for a file inside a Descent3 .pkg file."""

    def __init__(self, path: str) -> None:
        self.path = path
        self.offset = 0
        self.size = 0


class D3Pkg(SimpleUnpackable):
    """Object representing a Descent3 .pkg file."""

    def __init__(self, path: str) -> None:
        self.path = path
        self.n_files = 0
        self.files: list[D3Entry] = []
        self._parse_archive()

    def __enter__(self) -> D3Pkg:
        return self

    def __exit__(
        self,
        _et: type[BaseException] | None = None,
        _ev: BaseException | None = None,
        _tb: TracebackType | None = None
    ) -> None:
        pass

    def _parse_archive(self) -> None:
        reader = open(self.path, 'rb')
        archive_size = os.path.getsize(self.path)

        magic = int.from_bytes(reader.read(4), byteorder='little')
        if magic != PKG_MAGIC:
            raise ValueError(
                '"%s" is not a .pkg: magic number does not match' % self.path)

        self.n_files = int.from_bytes(reader.read(4), byteorder='little')
        logger.debug('%s contains %d files', self.path, self.n_files)

        nextFileStart = reader.tell()

        while nextFileStart < archive_size:
            reader.seek(nextFileStart)

            length = int.from_bytes(reader.read(4), byteorder='little')
            str_dirname = reader.read(length)[:-1].decode('windows-1252')
            str_dirname = str_dirname.replace("\\", "/")

            length = int.from_bytes(reader.read(4), byteorder='little')
            str_basename = reader.read(length)[:-1].decode('windows-1252')

            entry = D3Entry(os.path.join(str_dirname, str_basename))
            entry.size = int.from_bytes(reader.read(4), byteorder='little')
            logger.debug('Found %s - size: %d', entry.path, entry.size)

            # skip the next 8 bytes (checksum ?)
            reader.seek(8, 1)

            entry.offset = reader.tell()
            nextFileStart = entry.offset + entry.size

            self.files.append(entry)

        reader.close()

    def extractall(
        self,
        outdir: str,
        members: Collection[str] | None = None,
    ) -> None:
        # TODO: The members hint is ignored, we only support extracting
        # everything

        reader = open(self.path, 'rb')

        for entry in self.files:
            dest_file = os.path.join(outdir, entry.path)
            mkdir_p(os.path.dirname(dest_file))
            reader.seek(entry.offset)

            logger.debug('Extracting %s ...', dest_file)
            with open(dest_file, "wb") as f:
                f.write(reader.read(entry.size))
        reader.close()

    def printdir(self) -> None:
        for entry in self.files:
            print("%s @%d - size: %d" % (entry.path, entry.offset, entry.size))

    @property
    def format(self) -> str:
        return 'd3pkg'


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--output', '-o', help='extract to OUTPUT', default=None)
    parser.add_argument('pkg')
    args = parser.parse_args()

    archive = D3Pkg(args.pkg)

    if args.output:
        archive.extractall(args.output)
    else:
        archive.printdir()



#!/usr/bin/python3
#
# Copyright © 2023 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

# detect packages that are both generated by GDP
# and into official the archive

import glob
import os

import apt

srcdir = os.path.dirname(
    os.path.dirname(
        os.path.abspath(__file__)
    )
)


gdp: set[str] = set()

for yaml_file in sorted(glob.glob(os.path.join(srcdir, 'data', '*.yaml'))):
    with open(yaml_file, encoding='utf-8') as raw:
        for line in raw:
            if line.strip() == 'packages:':
                break
        for line in raw:
            if line == '\n':
                continue
            if line[0] != ' ':
                break
            if line[2] not in (' ', '#'):
                package = line.strip(':\n ')
                gdp.add(package)

archive: set[str] = set()

cache = apt.Cache()
for p in cache.keys():
    o = cache[p]
    if o.candidate and o.candidate.downloadable:
        archive.add(o.shortname)

print(sorted(archive & gdp))

# ['alex4-data', 'pushover-data']
# https://alioth-lists.debian.net/pipermail/pkg-games-devel/2023-December/063512.html
# https://alioth-lists.debian.net/pipermail/pkg-games-devel/2023-December/063510.html



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

# fake lgogdownloader to symlink in ~/bin

import argparse
import subprocess

print('FAKE LGOGDOWNLOADER:')

parser = argparse.ArgumentParser(prog='lgogdownloader',
                                 description='Fake lgogdownloader.')
parser.add_argument('--download', action='store_true')
parser.add_argument('--no-extra', action='store_true')
parser.add_argument('--directory', metavar='DIR')
parser.add_argument('--subdir-game', type=str)
parser.add_argument('--platform', type=str)
parser.add_argument('--include', type=str)
parser.add_argument('--language', type=str)
parser.add_argument('--game', type=str)
args = parser.parse_args()

assert args.directory, 'Must specify --directory'
assert args.game, 'Must specify --game'

game = args.game.lstrip('^').rstrip('$')

archives = {
  'descent#en': ['setup_descent_2.1.0.8.exe'],
  'legend_of_kyrandia#en': ['setup_legend_of_kyrandia_2.1.0.14.exe'],
  'legend_of_kyrandia#de': ['setup_legend_of_kyrandia_german_2.1.0.14.exe'],
  'legend_of_kyrandia#fr': ['setup_legend_of_kyrandia_french_2.1.0.14.exe'],
  'loom#en': ['setup_loom_2.0.0.4.exe'],
  'rise_of_the_triad__dark_war#en': ['setup_rise_of_the_triad_2.0.0.5.exe'],
  'space_quest_5_the_next_mutation': ['setup_space_quest5_2.1.0.15.exe'],
  'the_feeble_files#en': ['setup_the_feeble_files_2.0.0.5.exe',
                          'setup_the_feeble_files_2.0.0.5-1.bin',
                          'setup_the_feeble_files_2.0.0.5-2.bin'],
  'toonstruck#en': ['gog_toonstruck_2.0.0.4.sh'],
  'wolfenstein_3d#en': ['setup_wolfenstein3d_2.0.0.4.exe'],
  'wolfenstein_spear_of_destiny#en': ['setup_spear_of_destiny_2.0.0.6.exe'],
}.get(game + '#' + (args.language or 'en'))

if archives is None:
    exit('Unknown game %s' % game)

for archive in archives:
    try:
        locate = subprocess.check_output(['locate', '-e', archive], text=True)
    except subprocess.CalledProcessError:
        exit('Archive %s not found in "locate" database' % archive)
    for file in locate.splitlines():
        if file.endswith(archive):
            break
    else:
        exit('Archive %s not found in "locate" database' % archive)
    subprocess.check_call(['ln', '-s', '-v', file, args.directory])



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2024 Sébastien Noel <sebastien@twolife.be>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
import os
from typing import (TYPE_CHECKING)

from ..build import (PackagingTask)
from ..game import (GameData)
from ..util import (mkdir_p)

if TYPE_CHECKING:
    from typing import (Unpack)
    from ..packaging import (PerPackageState, PackagingTaskArgs)

logger = logging.getLogger(__name__)


class SC3UGameData(GameData):
    def construct_task(self, **kwargs: Unpack[PackagingTaskArgs]) -> SC3UShim:
        return SC3UShim(self, **kwargs)


class SC3UShim(PackagingTask):
    def fill_extra_files(
        self,
        per_package_state: PerPackageState
    ) -> None:
        super().fill_extra_files(per_package_state)
        package = per_package_state.package
        destdir = per_package_state.destdir

        if package.name == 'sc3u-data':
            # The game need these 2 empty directories
            gamedir = os.path.join(destdir, 'usr', 'share',
                                   'games', self.game.shortname)
            mkdir_p(os.path.join(gamedir, 'res/ba/ui/shared/dlg'))
            mkdir_p(os.path.join(gamedir, 'res/sound/radio/stations'))
            return


GAME_DATA_SUBCLASS = SC3UGameData



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import subprocess
import time

import distro_info
from debian.changelog import Changelog

# BASE = '/home/tchet/git'
# GDP = os.path.join(BASE, 'game-data-packager')

GDP = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
BASE = os.path.dirname(GDP)

# Support for DebHelper and DH-Cruft
DH_VERSIONS = {
    'xenial': (9, False),
    'jammy': (13, False),
    'kinetic': (13, False),
    'lunar': (13, True),
}

subprocess.check_call(['rm', '-rvf', 'ref.zip'],
                      cwd=GDP)
subprocess.check_call(['rm', '-rvf', 'build'],
                      cwd=GDP)
subprocess.check_call(['git', 'checkout', 'debian/changelog'],
                      cwd=GDP)
subprocess.check_call(['git', 'checkout', 'debian/control'],
                      cwd=GDP)

with open('debian/changelog', encoding='utf-8') as log:
    cl = Changelog(log, strict=False)

assert cl.distributions in ('unstable', 'UNRELEASED'), cl.distributions

if cl.distributions == 'unstable':
    build_type = 'Backport to PPA'
    build_number = 'release+'
else:
    build_type = 'Git snapshot'
    build_number = time.strftime('git%Y%m%d+')

lts = str(distro_info.UbuntuDistroInfo().lts())
current = str(distro_info.UbuntuDistroInfo().stable())

releases = sorted(set([lts, current]))
print('RELEASES:', releases)

with open('debian/control', 'r') as compat:
    for line in compat:
        if 'debhelper-compat' in line:
            current_debhelper = int(line.split('(')[1].strip(' =),\n'))
            break

for release in sorted(releases):
    supported_dh = DH_VERSIONS.get(release, [current_debhelper, False])
    supported_debhelper = supported_dh[0]
    supported_dh_cruft = supported_dh[1]
    backported = set()

    if supported_debhelper < current_debhelper:
        backported.add('control')
        build_dep = 'debhelper-compat ( = %d)' % supported_debhelper
        subprocess.check_call(['sed', '-i',
                               r's/\ *debhelper-compat.*/ ' + build_dep + ',/',
                               'debian/control'],
                              cwd=GDP)

    if not supported_dh_cruft:
        backported.add('control')
        subprocess.check_call(['sed', '-i', '/dh-sequence-cruft/d',
                              'debian/control'],
                              cwd=GDP)

    snapshot = str(cl.version) + '~' + build_number + release
    subprocess.check_call(['dch', '-b',
                           '-D', release,
                           '-v', snapshot,
                           build_type],
                          cwd=GDP)

    subprocess.check_call(['debuild', '-S', '-i'], cwd=GDP)
    subprocess.check_call(['dput', 'my-ppa',
                           'game-data-packager_%s_source.changes' % snapshot],
                          cwd=BASE)

    subprocess.check_call(['git', 'checkout', 'debian/changelog'],
                          cwd=GDP)

    for item in backported:
        subprocess.check_call(['git', 'checkout', 'debian/' + item],
                              cwd=GDP)
    for file in ('.tar.xz',
                 '.dsc',
                 '_source.build',
                 '_source.changes',
                 '_source.my-ppa.upload'):
        subprocess.check_call(['rm', '-v',
                               'game-data-packager_%s%s' % (snapshot, file)],
                              cwd=BASE)



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2024 Sébastien Noel
# SPDX-License-Identifier: GPL-2.0-or-later

import configparser
import json
import os
import sys
import zipfile

from contextlib import suppress
from typing import Any

from game_data_packager.data import Package
from game_data_packager.game import load_games
from game_data_packager.games.scummvm_common import ScummvmGameData
from game_data_packager.packaging import (
    PackagingSystem,
    get_native_packaging_system,
)
from game_data_packager.util import mkdir_p


def write_desktop_file(
        pgm: str,
        game: ScummvmGameData,
        package: Package,
        packaging: PackagingSystem,
        desktop_subdir: str,
) -> None:
    gdp_launcher = '$assets/game-data-packager-runtime/gdp-launcher'
    gdp_launcher = packaging.substitute(gdp_launcher, package.name)

    desktop = configparser.RawConfigParser()
    desktop.optionxform = lambda option: option     # type: ignore
    desktop['Desktop Entry'] = {}
    entry = desktop['Desktop Entry']
    entry['Name'] = package.longname or game.longname
    assert type(game.genre) is str
    entry['GenericName'] = game.genre + ' Game'
    entry['TryExec'] = 'scummvm'
    entry['Exec'] = '%s --id %s' % (gdp_launcher, pgm)
    entry['Icon'] = pgm
    entry['Terminal'] = 'false'
    entry['Type'] = 'Application'
    entry['Categories'] = (
        'Game;%sGame;' % game.genre.replace(' ', '')
    )

    with open(
        os.path.join(desktop_subdir, '%s.desktop' % pgm),
        'w', encoding='utf-8',
    ) as output:
        desktop.write(output, space_around_delimiters=False)


def main(vfs_zip: str, outfile: str, desktop_subdir: str) -> None:
    with suppress(FileNotFoundError):
        os.unlink(outfile)

    mkdir_p(desktop_subdir)
    packaging = get_native_packaging_system()
    games = load_games(use_vfs=vfs_zip)

    with zipfile.ZipFile(outfile, mode="w") as archive:
        for g_name, game in games.items():
            if not isinstance(game, ScummvmGameData):
                continue

            for p_name, package in game.packages.items():
                if package.section != 'games':
                    continue
                if package.data_type != 'data':
                    continue
                if package.name.endswith('-music'):
                    continue
                if package.name.endswith('-video'):
                    continue

                assert package.name.endswith('-data'), package.name
                pgm = package.name[0:len(package.name)-len('-data')]

                write_desktop_file(pgm, game, package,
                                   packaging, desktop_subdir)

                gameid = game.gameid
                if package.gameid is not None:
                    gameid = package.gameid

                langs = package.langs

                launcher_path = 'launch-%s.json' % pgm
                launcher: dict[str, Any] = {}
                launcher['argv'] = 'scummvm -p %s %s' % \
                                   (package.install_to, gameid)
                launcher['argv'] = packaging.substitute(
                                          launcher['argv'],
                                          package.name,
                                        )
                if len(langs) > 1 and gameid != 'nippon':
                    launcher['languages'] = {'choices': langs, 'arg': '-q'}

                info = zipfile.ZipInfo(launcher_path)
                archive.writestr(
                    info,
                    json.dumps(launcher),
                    compress_type=zipfile.ZIP_DEFLATED,
                    compresslevel=9,
                )


if __name__ == '__main__':
    main(*sys.argv[1:])



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2014-2016 Simon McVittie <smcv@debian.org>
# Copyright © 2015-2016 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
import os
import stat
import subprocess
from collections.abc import (Iterable)
from typing import (List, IO, TYPE_CHECKING)

try:
    from debian.deb822 import Deb822
    from debian.debian_support import Version
except ImportError:
    # make check
    Deb822 = None       # type: ignore
    Version = None      # type: ignore

from . import (Compression, PackagingSystem, PerPackageState)
from ..data import (HashedFile)
from ..util import (
        check_output,
        mkdir_p,
        normalize_permissions,
        rm_rf,
        run_as_root)

if TYPE_CHECKING:
    from ..data import (Package, PackageRelation)
    from ..game import (GameData)


logger = logging.getLogger(__name__)


class DebPackaging(PackagingSystem):
    BINDIR = '$prefix/games'
    ASSETS = '$datadir/games'
    CHECK_CMD = 'lintian'
    INSTALL_CMD = ['apt-get', 'install']
    # This is only used when cross-building Debian packages on
    # non-Debian, so conservatively assume that they don't have a new
    # enough dpkg-dev for --root-owner-group
    BUILD_DEP = {'dpkg', 'fakeroot', 'python3-debian'}
    PACKAGE_MAP = {
                  'id-shr-extract': 'dynamite',
                  'lha': 'lhasa',
                  '7z': '7zip',
                  'unrar-nonfree': 'unrar',
                  'zoom': 'zoom-player',
                  'doom': 'doom-engine',
                  'boom': 'boom-engine',
                  'heretic': 'heretic-engine',
                  'hexen': 'hexen-engine',
                  'doomsday-compat': 'doomsday',
                  }
    RENAME_PACKAGES = {
            'libSDL-1.2.so.0': 'libsdl1.2debian',
            'libSDL_mixer-1.2.so.0': 'libsdl-mixer1.2',
            'libSDL_ttf-2.0.so.0': 'libsdl-ttf2.0-0',
            'libgcc_s.so.1': 'libgcc1',
            'libjpeg.so.62': 'libjpeg62-turbo | libjpeg62',
            'libsmpeg-0.4.so.0': 'libsmpeg0t64 | libsmpeg0',
            'libz.so.1': 'zlib1g',
    }

    def __init__(self, architecture: str | None = None) -> None:
        super(DebPackaging, self).__init__(architecture=architecture)
        self.__installed: set[str] | None = None
        self.__available: set[str] | None = None
        self._contexts = ('deb', 'generic')

    def read_architecture(self) -> str:
        self._architecture = primary = check_output(
            ['dpkg', '--print-architecture']
        ).strip().decode('ascii')
        self._foreign_architectures = set(
            check_output(
                ['dpkg', '--print-foreign-architectures']
            ).strip().decode('ascii').split()
        )
        assert type(primary) is str
        return primary

    def is_installed(self, package: str) -> bool:
        # FIXME: this shouldn't be hard-coded
        if package == 'doom-engine':
            return (
                self.is_installed('chocolate-doom')
                or self.is_installed('crispy-doom')
                or self.is_installed('dsda-doom')
                or self.is_installed('woof-doom')
                or self.is_installed('doomsday')
            )
        if package == 'boom-engine':
            return (
                self.is_installed('dsda-doom')
                or self.is_installed('woof-doom')
                or self.is_installed('doomsday')
            )
        if package == 'heretic-engine':
            return (
                self.is_installed('chocolate-doom')
                or self.is_installed('crispy-doom')
                or self.is_installed('dsda-doom')
                or self.is_installed('doomsday')
            )
        if package == 'hexen-engine':
            return (
                self.is_installed('chocolate-hexen')
                or self.is_installed('crispy-doom')
                or self.is_installed('dsda-doom')
                or self.is_installed('doomsday')
            )

        if os.path.isdir(os.path.join('/usr/share/doc', package)):
            return True

        cache = self.__installed

        if cache is None:
            try:
                proc = subprocess.Popen(
                    ['dpkg-query', '--show', '--showformat', '${Package}\\n'],
                    text=True,
                    close_fds=True,
                    stdout=subprocess.PIPE,
                )
            except FileNotFoundError:
                return False

            cache = set()
            stdout = proc.stdout
            assert stdout is not None
            for line in stdout:
                cache.add(line.rstrip())
            self.__installed = cache
            stdout.close()
            proc.wait()

        return package in cache

    def is_available(self, package: str) -> bool:
        cache = self.__available

        if cache is None:
            try:
                proc = subprocess.Popen(
                    ['apt-cache', 'pkgnames'],
                    text=True,
                    close_fds=True,
                    stdout=subprocess.PIPE,
                )
            except FileNotFoundError:
                return False

            cache = set()
            stdout = proc.stdout
            assert stdout is not None
            for line in stdout:
                cache.add(line.rstrip())
            self.__available = cache
            stdout.close()
            proc.wait()

        return package in cache

    def current_version(self, package: str) -> str | None:
        # 'dpkg-query: no packages found matching $package'
        # will leak on stderr if called with an unknown package,
        # but that should never happen
        try:
            version = check_output(
                [
                    'dpkg-query', '--show', '--showformat', '${Version}',
                    package,
                ],
                text=True)
            assert type(version) is str
            return version
        except FileNotFoundError:
            return None
        except subprocess.CalledProcessError:
            return None

    def available_version(self, package: str) -> str | None:
        try:
            current_ver = check_output(
                ['apt-cache', 'madison', package], text=True,
            )
        except FileNotFoundError:
            return None
        assert type(current_ver) is str
        current_ver = current_ver.splitlines()[0]
        current_ver = current_ver.split('|')[1].strip()
        return current_ver

    def install_packages(
        self,
        debs: Iterable[str],
        method: str | None = 'apt',
        gain_root: str = 'su',
        force: bool = False,
    ) -> None:
        if method is None:
            method = 'apt'
        elif method not in (
            'apt', 'dpkg',
            'gdebi', 'gdebi-gtk', 'gdebi-kde',
        ):
            logger.warning(
                'Unknown installation method %r, using apt instead',
                method,
            )
            method = 'apt'

        if method == 'apt':
            argv = ['apt-get', 'install', '--install-recommends']

            if force:
                argv.append('--assume-yes')

            run_as_root(argv + list(debs), gain_root)
        elif method == 'dpkg':
            run_as_root(['dpkg', '-i'] + list(debs), gain_root)
        elif method == 'gdebi':
            run_as_root(['gdebi'] + list(debs), gain_root)
        else:
            # gdebi-gtk etc.
            subprocess.call([method] + list(debs))

    def rename_package(self, p: str) -> str:
        mapped = super(DebPackaging, self).rename_package(p)

        if mapped != p:
            return mapped

        p = p.lower().replace('_', '-')

        if '.so.' in p:
            lib, version = p.split('.so.', 1)

            if lib[-1] in '012345679':
                lib += '-'

            return lib + version

        return p

    def format_relation(self, pr: PackageRelation) -> str:
        assert not pr.contextual

        if pr.alternatives:
            return ' | '.join(
                [self.format_relation(p) for p in pr.alternatives]
            )

        package = pr.package
        assert package is not None

        if pr.version is not None:
            # foo (>= 1.0)
            return '%s (%s %s)' % (
                self.rename_package(package), pr.version_operator, pr.version,
            )

        return self.rename_package(package)

    def __generate_control(
        self,
        game: GameData,
        package: Package,
        destdir: str,
        component: str,
    ) -> Deb822:
        if Deb822 is None:
            raise FileNotFoundError(
                'Cannot generate .deb packages without python3-debian',
            )

        control = Deb822()
        control['Package'] = package.name
        control['Version'] = package.version
        control['Priority'] = 'optional'
        control['Maintainer'] = \
            'Debian Games Team <pkg-games-devel@lists.alioth.debian.org>'

        installed_size = 0
        # algorithm from https://bugs.debian.org/650077 designed to be
        # filesystem-independent
        for dirpath, dirnames, filenames in os.walk(destdir):
            if dirpath == destdir and 'DEBIAN' in dirnames:
                dirnames.remove('DEBIAN')
            # estimate 1 KiB per directory
            installed_size += len(dirnames)
            for f in filenames:
                stat_res = os.lstat(os.path.join(dirpath, f))
                if (stat.S_ISLNK(stat_res.st_mode) or
                        stat.S_ISREG(stat_res.st_mode)):
                    # take the real size and round up to next 1 KiB
                    installed_size += ((stat_res.st_size + 1023) // 1024)
                else:
                    # this will probably never happen in gdp, but assume
                    # 1 KiB per non-regular, non-directory, non-symlink file
                    installed_size += 1
        control['Installed-Size'] = str(installed_size)

        if component == 'main':
            control['Section'] = package.section
        else:
            control['Section'] = component + '/' + package.section

        if package.architecture == 'all':
            control['Architecture'] = 'all'
            if package.multi_arch is None:
                control['Multi-Arch'] = 'foreign'
        else:
            control['Architecture'] = self.get_architecture(
                    package.architecture)

        if (
            package.multi_arch is not None
            and package.multi_arch != 'no'
        ):
            control['Multi-Arch'] = package.multi_arch

        dep = dict()

        for rel in package.relations:
            if rel == 'build_depends':
                continue

            dep[rel] = self.merge_relations(package, rel)
            logger.debug('%s %s %s', package.name, rel, ', '.join(dep[rel]))

        if package.mutually_exclusive:
            dep['conflicts'] |= package.demo_for
            dep['conflicts'] |= package.better_versions

        if package.mutually_exclusive:
            dep['replaces'] |= dep['provides']

        engine = self.substitute(
                package.engine or game.engine,
                package.name)

        if engine:
            engine_noversion = []
            for e in engine.split(' | '):
                if '>=' in e:
                    e, ver = e.split(maxsplit=1)
                    ver = ver.strip('(>=) ')
                    dep['breaks'].add('%s (<< %s~)' % (e, ver))
                engine_noversion.append(e)
            engine = ' | '.join(engine_noversion)

        # We only 'recommends' & not 'depends'; to avoid
        # that GDP-generated packages get removed
        # if engine goes through some gcc/png/ffmpeg/... migration
        # and must be temporarily removed.
        # It's not like 'apt-get install ...' can revert this removal;
        # user may need to dig again for the original media....
        if package.engine:
            assert engine is not None
            dep['recommends'].add(engine)
        elif game.engine and (
            package.section != 'doc'
            and not package.expansion_for
        ):
            assert engine is not None
            dep['recommends'].add(engine)

        if package.expansion_for:
            # check if default heuristic has been overridden in yaml
            for p in dep['depends']:
                if package.expansion_for == p.split()[0]:
                    break
            else:
                dep['depends'].add(package.expansion_for)

        # dependencies derived from *other* package's data
        for other_package in game.packages.values():
            if other_package.expansion_for:
                if package.name == other_package.expansion_for:
                    dep['suggests'].add(other_package.name)
                else:
                    for relp in package.relations['provides']:
                        if relp.package == other_package.expansion_for:
                            dep['suggests'].add(other_package.name)

            if other_package.mutually_exclusive:
                if package.name in other_package.better_versions:
                    dep['replaces'].add(other_package.name)

                if package.name in other_package.demo_for:
                    dep['replaces'].add(other_package.name)

        # Shortcut: if A Replaces B, A automatically Conflicts B
        dep['conflicts'] |= dep['replaces']

        # keep only strongest dependency
        dep['recommends'] -= dep['depends']
        dep['suggests'] -= dep['recommends']
        dep['suggests'] -= dep['depends']

        for k, v in dep.items():
            if v:
                control[k.title()] = ', '.join(sorted(v))

        if 'Description' not in control:
            short_desc, long_desc = self.generate_description(game, package)
            control['Description'] = (
                short_desc + '\n '
                + '\n '.join([(line or '.') for line in long_desc]))

        return control

    def __fill_dest_dir_deb(
        self,
        game: GameData,
        per_package_state: PerPackageState,
    ) -> None:
        component = per_package_state.component
        destdir = per_package_state.destdir
        md5sums = per_package_state.md5sums
        package = per_package_state.package

        if component is None:
            component = package.component

        if component == 'local':
            per_package_state.lintian_overrides.add(
                'unknown-section local/{}'.format(
                    package.section,
                )
            )

        for o in sorted(per_package_state.lintian_overrides):
            lintiandir = os.path.join(destdir, 'usr/share/lintian/overrides')
            mkdir_p(lintiandir)

            with open(
                os.path.join(lintiandir, package.name),
                'a',
                encoding='utf-8'
            ) as writer:
                writer.write('%s: %s\n' % (package.name, o))

        # same output as in dh_md5sums
        if md5sums is None:
            md5sums = {}

        # we only compute here the md5 we don't have yet,
        # for the (small) GDP-generated files
        for dirpath, dirnames, filenames in os.walk(destdir):
            if os.path.basename(dirpath) == 'DEBIAN':
                continue
            for fn in filenames:
                full = os.path.join(dirpath, fn)
                if os.path.islink(full):
                    continue
                file = full[len(destdir)+1:]
                if file not in md5sums:
                    with open(full, 'rb') as opened:
                        hf = HashedFile.from_file(full, opened)
                        md5 = hf.md5
                        assert md5 is not None
                        md5sums[file] = md5

        debdir = os.path.join(destdir, 'DEBIAN')
        mkdir_p(debdir)
        md5sums_path = os.path.join(destdir, 'DEBIAN/md5sums')
        with open(md5sums_path, 'w', encoding='utf8') as outfile:
            for file in sorted(md5sums.keys()):
                outfile.write('%s  %s\n' % (md5sums[file], file))
        os.chmod(md5sums_path, 0o644)

        control = os.path.join(destdir, 'DEBIAN/control')
        with open(control, 'wb') as fd:
            self.__generate_control(game, package, destdir, component).dump(
                    fd=fd, encoding='utf-8')
        os.chmod(control, 0o644)

    def build_package(
        self,
        per_package_state: PerPackageState,
        game: GameData,
        destination: str,
        compress: Compression = True,
    ) -> str:
        if Version is None:
            raise FileNotFoundError(
                'Cannot generate .deb packages without python3-debian',
            )

        destdir = per_package_state.destdir
        package = per_package_state.package
        arch = self.get_effective_architecture(package)
        self.__fill_dest_dir_deb(game, per_package_state)
        normalize_permissions(destdir)

        # it had better have a /usr and a DEBIAN directory or
        # something has gone very wrong
        assert os.path.isdir(os.path.join(destdir, 'usr')), destdir
        assert os.path.isdir(os.path.join(destdir, 'DEBIAN')), destdir

        deb_basename = '%s_%s_%s.deb' % (package.name, package.version, arch)

        outfile = os.path.join(os.path.abspath(destination), deb_basename)

        if not compress:
            dpkg_deb_args = ['-Znone']
        elif compress is True:
            dpkg_deb_args = []
        elif isinstance(compress, str):
            dpkg_deb_args = ['-Z' + compress]
        elif isinstance(compress, list):
            dpkg_deb_args = compress[:]

        dpkg_deb_args.insert(0, 'dpkg-deb')

        dpkg_version = Version(self.current_version('dpkg'))

        if dpkg_version >= Version('1.19.0'):
            dpkg_deb_args.append('--root-owner-group')
        else:
            dpkg_deb_args.insert(0, 'fakeroot')

        try:
            logger.info('generating package %s', package.name)
            check_output(
                    dpkg_deb_args +
                    ['-b', 'DESTDIR', outfile],
                    cwd=per_package_state.per_package_dir)
        except subprocess.CalledProcessError as cpe:
            print(cpe.output)
            raise

        rm_rf(destdir)
        return outfile

    def available_version_at_least(
        self,
        package: str,
        desired: str,
        *,
        is_installed: bool | None = None,
    ) -> bool:
        if Version is None:
            # Stub: assume yes it is
            return True

        if is_installed is None:
            is_installed = self.is_installed(package)

        if is_installed:
            current_ver = self.current_version(package)
        else:
            current_ver = self.available_version(package)

        # We automatically add a '~' suffix so that ">= 1.0-1" in the YAML
        # really means ">= 1.0-1~" on Debian, while not breaking packaging
        # systems that don't support the RPM/dpkg meaning of '~'
        if current_ver and Version(current_ver) >= Version(desired + '~'):
            return True

        return False

    def _parse_table_file(self, fd: IO[str]) -> Iterable[List[str]]:
        for line in fd:
            line = line.rstrip()
            if not line or line.startswith("#"):
                continue
            yield line.split()

    def _load_dpkg_tables(
        self,
        path: str = '/usr/share/dpkg'
    ) -> tuple[
        # Debian name: (GNU name, config.guess regex, bits, endian)
        dict[str, tuple[str, str, str, str]],
        # Debian name: (GNU name, config.guess regex)
        dict[str, tuple[str, str]],
        # Debian architecture: (ABI, libc, kernel, CPU)
        dict[str, tuple[str, str, str, str]],
    ]:
        cputable_path = os.path.join(path, 'cputable')
        ostable_path = os.path.join(path, 'ostable')
        tupletable_path = os.path.join(path, 'tupletable')

        assert os.path.exists(cputable_path)
        assert os.path.exists(ostable_path)
        assert os.path.exists(tupletable_path)

        cpu_fd = open(cputable_path, encoding='utf-8')
        os_fd = open(ostable_path, encoding='utf-8')
        tuple_fd = open(tupletable_path, encoding='utf-8')

        cputable = {}
        for row in self._parse_table_file(cpu_fd):
            cputable[row[0]] = tuple(row[1:])
        cpu_fd.close()

        ostable = {}
        for row in self._parse_table_file(os_fd):
            ostable[row[0]] = tuple(row[1:])
        os_fd.close()

        cpu_list = list(cputable.keys())
        archtable = {}
        for row in self._parse_table_file(tuple_fd):
            dpkg_tuple = row[0]
            dpkg_arch = row[1]
            if '<cpu>' in dpkg_tuple:
                for cpu_name in cpu_list:
                    debtuple_cpu = dpkg_tuple.replace('<cpu>', cpu_name)
                    dpkg_arch_cpu = dpkg_arch.replace('<cpu>', cpu_name)
                    if dpkg_arch_cpu not in archtable:
                        archtable[dpkg_arch_cpu] = debtuple_cpu.split('-', 3)
            else:
                archtable[dpkg_arch] = dpkg_tuple.split('-', 3)
        tuple_fd.close()

        return (cputable, ostable, archtable)

    def _dpkg_arch_to_multiarch(self, arch: str) -> str:
        cputable, ostable, archtable = self._load_dpkg_tables()

        debtuple = archtable[arch]
        abi = debtuple[0]
        libc = debtuple[1]
        os = debtuple[2]
        cpu = debtuple[3]

        assert cpu in cputable
        assert f'{abi}-{libc}-{os}' in ostable
        gnutriplet = '-'.join((
                             cputable[cpu][0],
                             ostable[f'{abi}-{libc}-{os}'][0]
                         ))
        return gnutriplet.replace('i686', 'i386')

    def get_libdir(self, package_arch: str) -> str:
        multiarch = self._dpkg_arch_to_multiarch(
                        self.get_architecture(package_arch)
                    )
        return f'/usr/lib/{multiarch}'


def get_packaging_system(
    distro: str | None = None,
    architecture: str | None = None
) -> PackagingSystem:
    return DebPackaging(architecture=architecture)



#!/usr/bin/python3
# encoding=utf-8

# Copyright 2015-2022 Simon McVittie
# Copyright 2015-2016 Alexandre Detiste
# SPDX-License-Identifier: GPL-2.0-or-later

import os

if os.environ.get('GDP_UNINSTALLED'):
    _srcdir = os.path.dirname(
        os.path.dirname(
            os.path.abspath(__file__)
        )
    )

    CONFIG = os.path.join(_srcdir, 'etc', 'game-data-packager.conf')

    if 'GDP_PKGDATADIR' in os.environ:
        DATADIR = os.environ['GDP_PKGDATADIR']
    else:
        DATADIR = os.path.join(_srcdir, 'out')

    ETCDIR = os.path.join(_srcdir, 'etc')
    ICONS = os.path.join(DATADIR, 'icons')
else:
    CONFIG = '/etc/game-data-packager.conf'

    if 'GDP_PKGDATADIR' in os.environ:
        DATADIR = os.environ['GDP_PKGDATADIR']
    elif os.path.isdir('/usr/share/games/game-data-packager'):
        DATADIR = '/usr/share/games/game-data-packager'
    else:
        DATADIR = '/usr/share/game-data-packager'

    ETCDIR = '/etc/game-data-packager'
    ICONS = DATADIR



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import glob
import logging
import os
import tempfile
import xml.etree.ElementTree
import urllib.request
from collections.abc import (Iterator)
from typing import (TYPE_CHECKING)

from .build import (
    BinaryExecutablesNotAllowed,
    DownloadsFailed,
    NoPackagesPossible,
)
from .packaging import (get_native_packaging_system)
from .util import (
    AGENT,
    ascii_safe,
    lang_score,
    rm_rf,
)

if TYPE_CHECKING:
    import argparse
    from .game import GameData

logger = logging.getLogger(__name__)


def parse_acf(path: str) -> Iterator[dict[str, str]]:
    for manifest in glob.glob(path + '/*.acf'):
        with open(manifest) as data:
            # the .acf files are not really JSON files
            level = 0
            acf_struct = {}
            for line in data.readlines():
                if line.strip() == '{':
                    level += 1
                elif line.strip() == '}':
                    level -= 1
                elif level != 1:
                    continue
                elif '"\t\t"' in line:
                    key, value = line.split('\t\t')
                    key = key.strip().strip('"')
                    value = value.strip().strip('"')
                    if key in ('appid', 'name', 'installdir'):
                        acf_struct[key] = value
            if 'name' not in acf_struct:
                acf_struct['name'] = acf_struct['installdir']
            yield acf_struct


STEAM_GAMES: list[tuple[int, str]] | None = None


def owned_steam_games(
    steam_id: str | None = None,
) -> list[tuple[int, str]]:
    global STEAM_GAMES
    if steam_id is None:
        steam_id = get_steam_id()
    if STEAM_GAMES is not None:
        return STEAM_GAMES
    STEAM_GAMES = []
    if steam_id is None:
        return []
    url = "http://steamcommunity.com/profiles/" + steam_id + "/games?xml=1"
    try:
        html = urllib.request.urlopen(
            urllib.request.Request(url, headers={'User-Agent': AGENT}),
        )
        tree = xml.etree.ElementTree.ElementTree()
        tree.parse(html)
        games_xml = tree.iter('game')
        for game in games_xml:
            el = game.find('appID')
            assert el
            assert el.text
            appid = int(el.text)
            el = game.find('name')
            assert el
            assert el.text
            name = el.text
            # print(appid, name)
            STEAM_GAMES.append((appid, name))
    except urllib.error.URLError:
        # e="[Errno 111] Connection refused" but e.errno=None ?
        pass

    return STEAM_GAMES


def get_steam_vdf() -> str | None:
    for p in ['~/.steam/debian-installation/config/loginusers.vdf',
              '~/.steam/config/loginusers.vdf']:
        path = os.path.expanduser(p)
        if os.path.isfile(path):
            return path
    return None


def get_steam_id() -> str | None:
    path = get_steam_vdf()
    if path is None:
        return None
    with open(path, 'r', ) as data:
        for line in data.readlines():
            line = line.strip('\t\n "')
            if line not in ('users', '{'):
                return line
    return None


def get_steam_account() -> str | None:
    path = get_steam_vdf()
    if path is None:
        return None
    with open(path, 'r', ) as data:
        for line in data.readlines():
            if 'AccountName' in line:
                return line.split('"')[-2]
    return None


class FoundPackage:
    def __init__(
        self,
        game: str,
        game_type: int,
        package: str,
        installed: bool,
        longname: str,
        paths: list[str],
    ) -> None:
        self.game = game
        self.game_type = type
        self.package = package
        self.installed = installed
        self.longname = longname
        self.paths = paths


def run_steam_meta_mode(
    args: argparse.Namespace,
    games: dict[str, GameData]
) -> None:
    logger.info(
        'Visit our community page: '
        'https://steamcommunity.com/groups/debian_gdp#curation',
    )
    owned = set()
    if args.download:
        steam_id = get_steam_id()
        if steam_id is None:
            logger.error(
                "Couldn't read SteamID from ~/.steam/config/loginusers.vdf"
            )
        else:
            logger.info('Getting list of owned games from '
                        'http://steamcommunity.com/profiles/' + steam_id)
            owned = set(g[0] for g in owned_steam_games(steam_id))

    logging.info('Searching for locally installed Steam games...')
    found_games = []
    found_packages = []
    tasks = {}
    packaging = get_native_packaging_system()

    for game, gamedata in games.items():
        for package in gamedata.packages.values():
            id = package.steam.get('id') or gamedata.steam.get('id')
            if not id:
                continue

            if package.type == 'demo':
                continue
            # ignore other translations for "I Have No Mouth"
            if lang_score(package.lang) == 0:
                continue

            installed = packaging.is_installed(package.name)
            if args.new and installed:
                continue

            if game not in tasks:
                tasks[game] = gamedata.construct_task(packaging=packaging)

            paths = []
            for path in tasks[game].iter_steam_paths((package,)):
                if path not in paths:
                    paths.append(path)
            if not paths and id not in owned:
                continue

            if game not in found_games:
                found_games.append(game)
            found_packages.append(FoundPackage(
                game,
                1 if package.type == 'full' else 2,
                package.name,
                installed,
                package.longname or gamedata.longname,
                paths
            ))
    if not found_games:
        logger.error('No Steam games found')
        return

    print('[x] = package is already installed')
    print('-' * 70 + '\n')
    found_packages = sorted(
        found_packages,
        key=lambda k: (k.game, k.game_type, k.longname),
    )
    for g in sorted(found_games):
        print(g)
        for p in found_packages:
            if p.game != g:
                continue
            print(
                '[%s] %-42s    %s' % (
                    'x' if p.installed else ' ',
                    p.package,
                    ascii_safe(p.longname)
                )
            )
            for path in p.paths:
                print(path)
            if not p.paths:
                print('<game owned but not installed/found>')
        print()

    if not args.new and not args.all:
        logger.info(
            'Please specify --all or --new to create desired packages.'
        )
        return

    preserve = (getattr(args, 'destination', None) is not None)
    install = getattr(args, 'install', True)
    if getattr(args, 'compress', None) is None:
        # default to not compressing if we aren't going to install it
        # anyway
        args.compress = preserve

    all_packages: set[str] = set()

    for shortname in sorted(found_games):
        task = tasks[shortname]
        task.verbose = getattr(args, 'verbose', False)
        task.save_downloads = args.save_downloads
        try:
            task.look_for_files(binary_executables=args.binary_executables)
        except BinaryExecutablesNotAllowed:
            continue
        except NoPackagesPossible:
            continue

        todo = list()
        for found_package in found_packages:
            if found_package.game == shortname and found_package.paths:
                todo.append(task.game.packages[found_package.package])

        if not todo:
            continue

        try:
            ready = task.prepare_packages(log_immediately=False,
                                          packages=todo)
        except NoPackagesPossible:
            logger.error('No package possible for %s.' % task.game.shortname)
            continue
        except DownloadsFailed:
            logger.error('Unable to complete any packages of %s'
                         ' because downloads failed.' % task.game.shortname)
            continue

        if args.destination is None:
            destination = workdir = tempfile.mkdtemp(prefix='gdptmp.')
        else:
            workdir = None
            destination = args.destination

        debs = task.build_packages(
            ready,
            compress=getattr(args, 'compress', True),
            destination=destination,
            download=getattr(args, 'download', True),
        )
        rm_rf(os.path.join(task.get_workdir(), 'tmp'))

        if preserve:
            for deb in debs:
                print('generated "%s"' % os.path.abspath(deb))
        all_packages = all_packages.union(debs)

    if not all_packages:
        logger.error('Unable to package any game.')
        if workdir:
            rm_rf(workdir)
        raise SystemExit(1)

    if install:
        packaging.install_packages(
            all_packages, args.install_method, args.gain_root_command,
        )
    if workdir:
        rm_rf(workdir)



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Simon McVittie <smcv@debian.org>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
import os
import shutil
import subprocess
from collections.abc import (Iterable)
from typing import (Any, TYPE_CHECKING)

from ..build import (PackagingTask, NoPackagesPossible)
from ..game import (GameData)
from ..util import (mkdir_p)

if TYPE_CHECKING:
    import argparse
    from typing import (Unpack)
    from ..data import (Package)
    from ..packaging import (PerPackageState, PackagingTaskArgs)

logger = logging.getLogger(__name__)


class LGeneralGameData(GameData):
    def add_parser(
        self,
        parsers: argparse._SubParsersAction[Any],
        base_parser: argparse.ArgumentParser,
        **kwargs: Any,
    ) -> 'argparse.ArgumentParser':
        parser = super(LGeneralGameData, self).add_parser(parsers, base_parser)
        parser.add_argument(
            '-f', dest='download', action='store_false',
            help='Require pg-data.tar.gz on the command line',
        )
        parser.add_argument(
            '-w', dest='download', action='store_true',
            help='Download pg-data.tar.gz (done automatically if necessary)',
        )
        return parser

    def construct_task(
        self,
        **kwargs: Unpack[PackagingTaskArgs]
    ) -> LGeneralTask:
        return LGeneralTask(self, **kwargs)


class LGeneralTask(PackagingTask):
    def prepare_packages(
        self,
        packages: Iterable[Package] | None = None,
        build_demos: bool = False,
        download: bool = True,
        search: bool = True,
        log_immediately: bool = True,
        everything: bool = False,
        requested_packages: set[Package] = set()
    ) -> set[Package]:
        # don't bother even trying if it isn't going to work
        if shutil.which('lgc-pg') is None:
            logger.error('The "lgc-pg" tool is required for this package.')
            raise NoPackagesPossible()

        if 'DISPLAY' not in os.environ and 'WAYLAND_DISPLAY' not in os.environ:
            logger.error('The "lgc-pg" tool requires '
                         'to run in some graphical environment.')
            raise NoPackagesPossible()

        ready = super(LGeneralTask, self).prepare_packages(
            packages,
            build_demos=build_demos,
            download=download,
            search=search,
            log_immediately=log_immediately,
            everything=everything,
            requested_packages=requested_packages,
        )

        # would have raised an exception if not
        assert self.game.packages['lgeneral-data-nonfree'] in ready
        return ready

    def fill_extra_files(
        self,
        per_package_state: PerPackageState
    ) -> None:
        super().fill_extra_files(per_package_state)
        package = per_package_state.package
        destdir = per_package_state.destdir
        assert package.name == 'lgeneral-data-nonfree'

        installdir = os.path.join(destdir, 'usr/share/games/lgeneral')
        unpackdir = os.path.join(self.get_workdir(), 'tmp', 'pg-data.tar.gz.d')

        for d in (
            'gfx/flags', 'gfx/terrain', 'gfx/units', 'maps/pg',
            'nations', 'scenarios/pg', 'sounds/pg', 'units',
        ):
            mkdir_p(os.path.join(installdir, d))

        shutil.unpack_archive(
                os.path.join(installdir, 'pg-data.tar.gz'),
                unpackdir,
                'gztar')

        subprocess.check_call([
            'lgc-pg', '-s', unpackdir + '/pg-data/',
            '-d', installdir + '/',
        ])


GAME_DATA_SUBCLASS = LGeneralGameData



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2022 Simon McVittie
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import subprocess
import sys
from contextlib import suppress


def main(zipfile: str, *inputs: str) -> None:
    os.environ['TZ'] = 'UTC'

    if 'SOURCE_DATE_EPOCH' in os.environ:
        timestamp = int(os.environ['SOURCE_DATE_EPOCH'])
    else:
        timestamp = None

    files = []

    for json in inputs:
        for ext in [
            'files',
            'groups',
            'json',
            'sha1sums',
            'sha256sums',
            'size_and_md5',
        ]:
            other = os.path.splitext(json)[0] + '.' + ext

            if os.path.exists(other):
                os.chmod(other, 0o644)

                if timestamp is not None:
                    os.utime(other, (timestamp, timestamp))

                files.append(other.encode('utf-8'))

    with suppress(FileNotFoundError):
        os.unlink(zipfile)

    # TODO: We could do this in pure Python too...
    subprocess.run(
        ['zip', zipfile, '-9', '-X', '-q', '-j', '-@'],
        check=True,
        input=b'\n'.join(sorted(files)) + b'\n',
    )


if __name__ == '__main__':
    main(*sys.argv[1:])



# encoding=utf-8

# Copyright 2015-2022 Simon McVittie
# Copyright 2015-2016 Alexandre Detiste
# SPDX-License-Identifier: GPL-2.0-or-later

# This version of this file is only used when run uninstalled. It is replaced
# with a generated version during installation.

import os

srcdir = os.path.dirname(
    os.path.dirname(
        os.path.abspath(__file__)
    )
)

with open(os.path.join(srcdir, 'debian', 'changelog'), encoding='utf-8') as fd:
    try:
        from debian.changelog import Changelog
    except ImportError:
        GAME_PACKAGE_VERSION = fd.readline().split('(', 1)[1].split(')', 1)[0]
        if '-' in GAME_PACKAGE_VERSION:
            GAME_PACKAGE_VERSION, GAME_PACKAGE_RELEASE = (
                GAME_PACKAGE_VERSION.rsplit('-', 1)
            )
        else:
            GAME_PACKAGE_RELEASE = ''
    else:
        cl = Changelog(fd, strict=False, max_blocks=1)
        GAME_PACKAGE_VERSION = cl.upstream_version or ''
        GAME_PACKAGE_RELEASE = cl.debian_revision or ''

details = {}
if os.path.isfile('/etc/os-release'):
    with open('/etc/os-release', encoding='utf-8') as release:
        for line in release:
            if '=' not in line:
                continue
            key, value = line.strip().split('=', 1)
            details[key] = value.strip('"')

if os.path.isfile('/etc/debian_version'):
    FORMAT = 'deb'
    DISTRO = 'generic'

# mageia also has a /etc/redhat-release
elif os.path.isfile('/etc/mageia-release'):
    FORMAT = 'rpm'
    DISTRO = 'mageia'

elif os.path.isfile('/etc/redhat-release'):
    FORMAT = 'rpm'
    DISTRO = 'fedora'

elif os.path.isfile('/etc/SuSE-release') or details.get('ID_LIKE') == 'suse':
    FORMAT = 'rpm'
    DISTRO = 'suse'

elif os.path.isfile('/etc/arch-release'):
    FORMAT = 'arch'
    DISTRO = 'arch'

else:
    exit('ERROR: Unknown distribution')

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--distro', default='')
    parser.add_argument('--format', default='')
    parser.add_argument('--version', default='')
    parser.add_argument('--release', default='')
    args = parser.parse_args()

    if args.release:
        GAME_PACKAGE_RELEASE = args.release

    if args.version:
        GAME_PACKAGE_VERSION = args.version

    if args.format:
        FORMAT = args.format

    if args.distro:
        DISTRO = args.distro

    print('# Generated file, do not edit')
    for const in (
        'DISTRO',
        'FORMAT',
        'GAME_PACKAGE_VERSION',
        'GAME_PACKAGE_RELEASE',
    ):
        print('%s = %r' % (const, eval(const)))



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

# a small tool to create or verify a mirror
# of downloadable files needed by GDP

# call it this way:
#   GDP_UNINSTALLED=1 python3 -m tools.mirror

# this tool will never remove any bad file,
# you'll have to investigate by yourself

import argparse
import os
import urllib.request

from game_data_packager.command_line import (TerminalProgress)
from game_data_packager.data import (HashedFile)
from game_data_packager.download import (Downloader)
from game_data_packager.game import (load_games)
from game_data_packager.util import (AGENT)

KEEP_FREE_SPACE = 250 * 1024 * 1024


class Archive:
    filename: str
    name: str
    size: int
    md5: str | None
    sha1: str | None
    sha256: str | None
    download: str


os.environ.pop('GDP_MIRROR', None)

parser = argparse.ArgumentParser()
parser.add_argument('--destination', default='/var/www/html')
args = parser.parse_args()

print('loading game definitions...')


def load_archives() -> list[Archive]:
    archives: list[Archive] = []

    for gamename, game in load_games().items():
        game.load_file_data()
        for filename, file in game.files.items():
            if file.unsuitable:
                # quake2-rogue-2.00.tar.xz could have been tagged this way
                fullname = os.path.join(args.destination, filename)
                if os.path.isfile(fullname):
                    print('Obsolete archive: %s (%s)' %
                          (fullname, file.unsuitable))
                continue
            elif not file.download:
                continue

            url = Downloader.choose_mirror(file)[0]
            if '?' not in url:
                destname = os.path.basename(url)
            elif '?' not in filename:
                destname = filename
            elif url.endswith('?download'):
                destname = os.path.basename(url)
                destname = destname[0:len(destname)-len('?download')]
            else:
                print("Can't compute filename for %s = %s" % (filename, url))
                # bothersome hexen2demo_nov1997-linux-i586.tgz
                continue

            assert file.size, filename
            assert file.md5 or file.sha1 or file.sha256, filename

            archive = Archive()
            archive.filename = filename
            archive.name = destname
            archive.size = file.size
            archive.md5 = file.md5
            archive.sha1 = file.sha1
            archive.sha256 = file.sha256
            archive.download = url
            archives.append(archive)
    return archives


archives = load_archives()
archives = sorted(archives, key=lambda k: (k.size))

seen: set[str] = set()

for a in archives:
    # always prefer the smallest version of a file (e.g.: tnt31fix.zip)
    if a.name in seen:
        continue
    seen.add(a.name)

    archive = os.path.join(args.destination, a.name)

    if not os.path.isfile(archive):
        statvfs = os.statvfs(args.destination)
        freespace = statvfs.f_frsize * statvfs.f_bavail
        if a.size > freespace - KEEP_FREE_SPACE:
            print('out of space, can not download %s' % a.name)
            continue

        try:
            rf = urllib.request.urlopen(urllib.request.Request(a.download,
                                        headers={'User-Agent': AGENT}))
        except urllib.error.HTTPError as e:
            print(e, a.name, a.download)
            continue

        if rf is None:
            continue

        with open(archive, 'wb') as wf:
            hf = HashedFile.from_file(a.download, rf, wf, size=a.size,
                                      progress=TerminalProgress(
                                        info='downloading %s as %s' %
                                        (a.download, a.name)))
        rf.close()
    else:
        print('checking %s ...' % archive)
        hf = HashedFile.from_file(archive,
                                  open(archive, 'rb'),
                                  size=a.size,
                                  progress=TerminalProgress())

    if os.path.getsize(archive) == 0:
        exit("%s is empty !!!" % archive)
    if os.path.getsize(archive) != a.size:
        exit("%s has the wrong size !!!" % archive)
    if a.md5 and a.md5 != hf.md5:
        exit("md5 doesn't match for %s !!!" % archive)
    if a.sha1 and a.sha1 != hf.sha1:
        exit("sha1 doesn't match for %s !!!" % archive)
    if a.sha256 and a.sha256 != hf.sha256:
        exit("sha256 doesn't match for %s !!!" % archive)



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2016 Simon McVittie <smcv@debian.org>
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import sys
import unittest

if 'GDP_UNINSTALLED' not in os.environ:
    sys.path.insert(0, '/usr/share/game-data-packager')
    sys.path.insert(0, '/usr/share/games/game-data-packager')

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))

from game_data_packager.data import (PackageRelation)  # noqa: E402
from game_data_packager.packaging.rpm import (RpmPackaging)  # noqa: E402


class RpmTestCase(unittest.TestCase):
    def setUp(self) -> None:
        pass

    def test_relation(self) -> None:
        # a random distribution that we don't actually support
        rp = RpmPackaging('caldera')

        def t(
            in_: list[str | dict[str, str]],
            out: list[str]
        ) -> None:
            self.assertEqual(
                    sorted(rp.format_relations(map(PackageRelation, in_))),
                    out)

        t(['libc.so.6'], ['libc.so.6'])
        t(['libc.so.6 (>= 2.19)'], ['libc.so.6 >= 2.19'])
        t(['libjpeg.so.62'], ['libjpeg.so.62'])
        t(['libc.so.6', 'libopenal.so.1'], ['libc.so.6', 'libopenal.so.1'])
        t([dict(deb='foo', rpm='bar')], ['bar'])
        t([dict(deb='foo', rpm='bar', generic='baz')], ['bar'])
        t([dict(deb='foo', generic='baz')], ['baz'])
        t([dict(deb='foo')], [])
        t([dict(rpm='foo', caldera='bar', fedora='baz')], ['bar'])

        with self.assertRaises(AssertionError):
            rp.format_relation(
                    PackageRelation('libopenal.so.1 | bundled-openal'))

    def tearDown(self) -> None:
        pass


if __name__ == '__main__':
    from gdp_test_common import main
    main()



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2014-2015 Simon McVittie <smcv@debian.org>
# Copyright © 2015-2016 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

import grp
import logging
import os
import shlex
import shutil
import stat
import subprocess
import sys
import yaml
from types import TracebackType
from typing import (Any, TextIO, Type)

SafeLoader: Type[yaml.CSafeLoader] | Type[yaml.SafeLoader]

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader

from .paths import DATADIR  # noqa: E402
from .version import (GAME_PACKAGE_VERSION, GAME_PACKAGE_RELEASE)  # noqa: E402

logger = logging.getLogger(__name__)

KIBIBYTE = 1024
MEBIBYTE = KIBIBYTE * KIBIBYTE

AGENT = (
    'Debian Game-Data-Packager/%s%s (%s %s;'
    ' +http://wiki.debian.org/Games/GameDataPackager)'
    % (
        GAME_PACKAGE_VERSION,
        GAME_PACKAGE_RELEASE,
        os.uname()[0],
        os.uname()[4],
    )
)


class TemporaryUmask:
    """Context manager to set the umask. Not thread-safe.

    Use like this::

        with TemporaryUmask(0o022):
            os.makedirs('usr/share/misc')
    """

    def __init__(self, desired_mask: int) -> None:
        self.desired_mask = desired_mask
        self.saved_mask: int | None = None

    def __enter__(self) -> None:
        self.saved_mask = os.umask(self.desired_mask)

    def __exit__(
        self,
        _et: type[BaseException] | None = None,
        _ev: BaseException | None = None,
        _tb: TracebackType | None = None
    ) -> None:
        saved_mask = self.saved_mask

        if saved_mask is not None:
            os.umask(saved_mask)


def load_yaml(stream: TextIO | str) -> Any:
    return yaml.load(stream, Loader=SafeLoader)


def mkdir_p(path: str) -> None:
    if not os.path.isdir(path):
        with TemporaryUmask(0o022):
            os.makedirs(path)


def rm_rf(path: str) -> None:
    if os.path.exists(path):
        shutil.rmtree(path)


def human_size(size: int) -> str:
    # 0.0 KiB up to 1024.0 KiB
    if size < MEBIBYTE:
        return '%.1f KiB' % (size / KIBIBYTE)

    # 1.0 MiB or more
    return '%.1f MiB' % (size / (MEBIBYTE))


def copy_with_substitutions(from_: TextIO, to: TextIO, **kwargs: str) -> None:
    for line in from_:
        for k, v in kwargs.items():
            line = line.replace(k, v)
        to.write(line)


__cached_prefered_langs: list[str] | None = None


# Keep in sync with runtime/gdp_launcher_base.py
def prefered_langs() -> list[str]:
    global __cached_prefered_langs

    if __cached_prefered_langs is not None:
        return __cached_prefered_langs

    lang_raw: list[str] = []
    if 'LANGUAGE' in os.environ:
        lang_raw = os.environ['LANGUAGE'].split(':')
    if 'LANG' in os.environ:
        lang_raw.append(os.environ['LANG'])
    lang_raw.append('en')

    __cached_prefered_langs = []
    for lang in lang_raw:
        lang = lang.split('.')[0]
        if not lang or lang == 'C':
            continue
        if lang not in __cached_prefered_langs:
            __cached_prefered_langs.append(lang)
        lang = lang[0:2]
        if lang not in __cached_prefered_langs:
            __cached_prefered_langs.append(lang)

    return __cached_prefered_langs


def lang_score(lang: str) -> int:
    langs = prefered_langs()

    if lang in langs:
        return len(langs) - langs.index(lang)

    lang = lang[0:2]
    if lang in langs:
        return len(langs) - langs.index(lang)

    return 0


def ascii_safe(string: str, force: bool = False) -> str:
    if sys.stdout.encoding != 'UTF-8' or force:
        string = string.translate(
            str.maketrans(
                'àäçčéèêëîïíôöōłñù§┏┛',
                'aacceeeeiiiooolnu***',
            )
        )
    return string


def run_as_root(argv: list[str], gain_root: str = '') -> None:
    if not gain_root and shutil.which('pkexec') is not None:
        # Use pkexec if possible. It has desktop integration, and will
        # prompt for the user's password if they are administratively
        # privileged (a member of group sudo), or root's password
        # otherwise.
        gain_root = 'pkexec'

    if not gain_root and shutil.which('sudo') is not None:
        # Use sudo as the next choice after pkexec, but only if we're in
        # a group that should be able to use it.
        try:
            sudo_group = grp.getgrnam('sudo')
        except KeyError:
            pass
        else:
            if sudo_group.gr_gid in os.getgroups():
                gain_root = 'sudo'

        # If we are in the admin group, also use sudo, but only
        # if this looks like Ubuntu. We use dpkg-vendor at build time
        # to detect Ubuntu derivatives.
        try:
            admin_group = grp.getgrnam('admin')
        except KeyError:
            pass
        else:
            if (admin_group.gr_gid in os.getgroups() and
                    os.path.exists(
                        os.path.join(DATADIR, 'is-ubuntu-derived'))):
                gain_root = 'sudo'

    if not gain_root:
        # Use su if we don't have anything better.
        gain_root = 'su'

    if gain_root not in ('su', 'pkexec', 'sudo', 'super', 'really'):
        logger.warning(
            'Unknown privilege escalation method %r, assuming it works '
            'like sudo',
            gain_root)

    if gain_root == 'su':
        print('using su to obtain root privileges and install the package(s)')

        # su expects a single sh(1) command-line
        cmd = ' '.join([shlex.quote(arg) for arg in argv])

        subprocess.call(['su', '-c', cmd])
    else:
        # this code path works for pkexec, sudo, super, really;
        # we assume everything else is the same
        print(
            'using %s to obtain root privileges and install the package(s)' %
            gain_root)
        subprocess.call([gain_root] + list(argv))


def check_call(command: list[str], *args: Any, **kwargs: Any) -> None:
    """Like subprocess.check_call, but log what we will do first."""
    logger.debug('%r', command)
    subprocess.check_call(command, *args, **kwargs)


def check_output(command: list[str], *args: Any, **kwargs: Any) -> Any:
    """Like subprocess.check_output, but log what we will do first."""
    logger.debug('%r', command)
    return subprocess.check_output(command, *args, **kwargs)


def recursive_utime(
    directory: str,
    orig_time: float | tuple[float, float],
) -> None:
    """Recursively set the access and modification times of everything
    in directory to orig_time.

    orig_time may be a tuple (atime, mtime), or a single int or float.
    """
    if isinstance(orig_time, (int, float)):
        orig_time = (orig_time, orig_time)

    for dirpath, dirnames, filenames in os.walk(directory):
        for fn in filenames:
            full = os.path.join(dirpath, fn)
            os.utime(full, orig_time)


def normalize_permissions(directory: str) -> None:
    for dirpath, dirnames, filenames in os.walk(directory):
        for fn in filenames + dirnames:
            full = os.path.join(dirpath, fn)
            stat_res = os.lstat(full)
            if stat.S_ISLNK(stat_res.st_mode):
                continue
            elif stat.S_ISDIR(stat_res.st_mode):
                # make directories rwxr-xr-x
                os.chmod(full, 0o755)
            elif (stat.S_IMODE(stat_res.st_mode) & 0o111) != 0:
                # make executable files rwxr-xr-x
                os.chmod(full, 0o755)
            else:
                # make other files rw-r--r--
                os.chmod(full, 0o644)



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2014 Simon McVittie <smcv@debian.org>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from ..game import (GameData)

if TYPE_CHECKING:
    import argparse

logger = logging.getLogger(__name__)


class Wolf3DGameData(GameData):
    def add_parser(
        self,
        parsers: argparse._SubParsersAction[Any],
        base_parser: argparse.ArgumentParser,
        **kwargs: Any
    ) -> argparse.ArgumentParser:
        parser = super(Wolf3DGameData, self).add_parser(parsers, base_parser)
        parser.add_argument(
            '-f', dest='download', action='store_false',
            help='Require 1wolf14.zip on the command line',
        )
        parser.add_argument(
            '-w', dest='download', action='store_true',
            help='Download 1wolf14.zip (done automatically if necessary)',
        )
        return parser


GAME_DATA_SUBCLASS = Wolf3DGameData



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

from game_data_packager.game import (load_games)
from game_data_packager.packaging.deb import (DebPackaging)


class Package:
    game: str
    year: int
    type: str
    fanmade_: bool
    package: str
    disks: int = 1
    engine: str = ''
    size_min: int
    size_max: int

    def fanmade(self) -> str:
        return {True: 'Y'}.get(game.fanmade, 'N')

    def order(self) -> int:
        return {
            'demo': 1,
            'full': 2,
            'expansion': 3,
            'doc': 4
        }.get(self.type, 5)


games: list[Package] = []
packaging = DebPackaging()

for name, game in load_games().items():
    game.load_file_data()
    for package in game.packages.values():
        if package.rip_cd:
            continue
        p = Package()
        p.game = name
        p.package = package.name
        p.type = package.type

        engine = package.engine or game.engine or ''
        if type(engine) is dict:
            engine = packaging.substitute(engine, name)
        assert type(engine) is str
        p.engine = engine.split(' ')[0]

        copyright = package.copyright or game.copyright
        assert type(copyright) is str
        p.year = int(copyright.split()[1][0:4])

        p.size_min, p.size_max = game.size(package)
        p.fanmade_ = game.fanmade
        games.append(p)

games = sorted(games,
               key=lambda k: (k.game, k.order(), k.package))

print('GAME;YEAR;TYPE;FANMADE;PACKAGE;DISKS;ENGINE;SIZE_MIN;SIZE_MAX')
for g in games:
    print('%s;%d;%s;%s;%s;%d;%s;%d;%d' %
          (g.game, g.year, g.type, g.fanmade(), g.package,
           g.disks, g.engine, g.size_min, g.size_max))



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2015 Simon McVittie <smcv@debian.org>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import glob
import logging
import os
import subprocess
import tarfile
from typing import (TYPE_CHECKING)
from zipfile import ZipFile

from ..build import (FillResult, PackagingTask)
from ..game import (GameData)

if TYPE_CHECKING:
    from typing import (Unpack)
    from ..data import (Package)
    from ..packaging import (PerPackageState, PackagingTaskArgs)

logger = logging.getLogger(__name__)


class Quake2GameData(GameData):
    def construct_task(
        self,
        **kwargs: Unpack[PackagingTaskArgs]
    ) -> Quake2Task:
        return Quake2Task(self, **kwargs)


class Quake2Task(PackagingTask):
    def fill_gaps(
        self,
        package: Package | None,
        download: bool = False,
        log: bool = True,
        recheck: bool = False,
        requested: bool = False,
    ) -> FillResult:
        assert package is not None
        if package.name in (
            'quake2-reckoning-data',
            'quake2-groundzero-data',
            'quake2-zaero-data',
        ) and (
            self.packaging.get_architecture()
            != self.builder_packaging.get_architecture()
        ):
            logger.warning('Cross-compiling %s not supported', package.name)
            return FillResult.IMPOSSIBLE

        return super().fill_gaps(
            package,
            download=download,
            log=log,
            recheck=recheck,
            requested=requested,
        )

    def fill_extra_files(
        self,
        per_package_state: PerPackageState
    ) -> None:
        super().fill_extra_files(per_package_state)
        package = per_package_state.package
        destdir = per_package_state.destdir

        if package.name not in ('quake2-reckoning-data',
                                'quake2-groundzero-data',
                                'quake2-zaero-data',
                                ):
            return

        subdir = {
            'quake2-groundzero-data': 'rogue',
            'quake2-reckoning-data': 'xatrix',
            'quake2-zaero-data': 'zaero',
        }[package.name]

        source_is_github_commit_zip = {
            'quake2-groundzero-data': False,
            'quake2-reckoning-data': False,
            'quake2-zaero-data': True,
        }[package.name]

        installdir = os.path.join(destdir, 'usr', 'share', 'games', 'quake2')
        unpackdir = os.path.join(
            self.get_workdir(), 'tmp', package.name + '.build.d',
        )

        if source_is_github_commit_zip:
            zips = list(glob.glob(os.path.join(installdir, '*.zip')))
            assert len(zips) == 1, zips
            with ZipFile(zips[0]) as zObject:
                zObject.extractall(path=unpackdir)

            expect_dir_list = list(glob.glob(os.path.join(unpackdir, '*')))
            assert len(expect_dir_list) == 1, expect_dir_list
            expect_dir = os.path.basename(expect_dir_list[0])
        else:
            tars = list(glob.glob(os.path.join(installdir, '*.tar.xz')))
            assert len(tars) == 1, tars
            expect_dir = os.path.basename(tars[0])
            expect_dir = expect_dir[:len(expect_dir) - 7]

            with tarfile.open(tars[0], mode='r:xz') as tar:
                tar.extractall(unpackdir)

        subprocess.check_call([
            'make', '-C',
            os.path.join(unpackdir, expect_dir), '-s', '-j5',
        ])
        subprocess.check_call([
            'install', '-s', '-m644',
            os.path.join(unpackdir, expect_dir, 'release', 'game.so'),
            os.path.join(installdir, subdir, 'game.so'),
        ])


GAME_DATA_SUBCLASS = Quake2GameData



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2022 Simon McVittie
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import subprocess
import sys
import tempfile


def main(source: str, layer: str, output: str) -> None:
    with tempfile.TemporaryDirectory() as temp:
        subprocess.run(
            [
                'xmlstarlet',
                'ed',
                '-d',
                ("//*[local-name() = 'g' and @inkscape:groupmode = 'layer' "
                 "and @id != '{}']".format(layer)),
            ],
            check=True,
            stdin=open(source, 'r'),
            stdout=open(os.path.join(temp, 'layer.svg'), 'w'),
        )
        subprocess.run(
            [
                'inkscape',
                '--export-area-page',
                '--export-plain-svg',
                '--export-filename=' + output,
                os.path.join(temp, 'layer.svg'),
            ],
            check=True,
            stdin=open(os.path.join(temp, 'layer.svg'), 'r'),
        )


if __name__ == '__main__':
    main(sys.argv[1], sys.argv[2], sys.argv[3])



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2016 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

import hashlib
import sys

CDROM_SECTOR = 2048

if len(sys.argv) != 2:
    exit('Usage: cdrom_md5.py <file>')

with open(sys.argv[1], 'rb') as f:
    while True:
        blob = f.read(CDROM_SECTOR)
        if not blob:
            break
        md5 = hashlib.new('md5')
        md5.update(blob)
        print(md5.hexdigest())



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2016 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

import glob
import gzip
import os
import sys

scope: set[str] = set()
with open(
    os.path.join(os.environ.get('GDP_BUILDDIR', 'out'), 'bash_completion'),
    'r',
) as bash:
    # keep each even line
    while bash.readline():
        scope |= set(bash.readline().rstrip().split())

# old package names
scope.add('quake2-data')
scope.add('wolf3d-data-wl1')
scope.add('wolf3d-v14-data')
scope.add('rtcw-data')

# add this to compare against other Doom packages
scope.add('doom-wad-shareware')

scope.add('game-data-packager')
scope.add('game-data-packager-runtime')


def process(scope: set[str], distro: str) -> dict[str, int]:
    result: dict[str, int] = dict()
    if len(sys.argv) == 2:
        popcon = '/var/cache/popcon/' + distro + '_' + sys.argv[1] + '.gz'
    else:
        popcon = sorted(glob.glob('/var/cache/popcon/'
                                  + distro + '_*.gz')).pop()

    with gzip.open(popcon, 'rb') as f:
        file_content = f.read().decode('latin1')
    for line in file_content.split('\n'):
        if not line or line[0] in ('#', '-'):
            continue
        try:
            package, _score = line.split()[1:3]
            score = int(_score)
            if package in scope:
                # print("%30s %d" % (package, score))
                result[package] = score
        except ValueError:
            # print(distro,line)
            pass

    return result


debian: dict[str, int] = process(scope, 'debian')


class Game:
    name: str
    score: int

    def __init__(self, name: str, score: int) -> None:
        self.name = name
        self.score = score


games: list[Game] = []
for key in scope:
    game = Game(key, debian.get(key, 0))
    games.append(game)

games = sorted(games, key=lambda k: (-k.score, k.name))

print('Package                                             Debian')
print('------------------------------------------------------------')
for package in games:
    print('%49s %8d' % (package.name, package.score))



#!/usr/bin/python3
# encoding=utf-8

# game-data-packager Gtk launcher stub. See doc/launcher.mdwn for design

# Copyright © 2015-2016 Simon McVittie <smcv@debian.org>
# SPDX-License-Identifier: GPL-2.0-or-later

import enum
import fnmatch
import logging
import os
import sys
import traceback
from contextlib import suppress
from typing import (Any, TextIO)

import gi

gi.require_version('Gtk', '4.0')

from gi.repository import (GLib, GObject)       # noqa
from gi.repository import Gtk                   # noqa

from gdp_launcher_base import (
    Launcher,
    RUNTIME_BUILT,
    prefered_langs,
)    # noqa

logger = logging.getLogger('game-data-packager.launcher')

if os.environ.get('GDP_DEBUG'):
    logger.setLevel(logging.DEBUG)
else:
    logger.setLevel(logging.INFO)


class LanguageCodes(enum.StrEnum):
    cz = 'Czech'
    de = 'German'
    en = 'English'
    es = 'Spanish'
    fr = 'French'
    it = 'Italian'
    nl = 'Dutch'
    pl = 'Polish'
    ru = 'Russian'


class IniEditor:
    def __init__(
        self,
        edits: list[dict[str, str | list[str] | dict[str, str]]]
    ) -> None:
        self.lines: list[str] = []
        self.edits = edits
        self.__section: str | None = None
        self.__section_lines: list[str] = []
        self.__sections: set[str] = set()

    def load(self, reader: TextIO) -> None:
        # Simple INI parser. Not using ConfigParser because Unreal
        # uses duplicate keys within sections, and we want to preserve
        # comments, blank lines etc.
        self.__section = None
        self.__section_lines = []
        self.__sections = set()

        for line in reader:
            line = line.rstrip('\r\n')

            if line.startswith('[') and line.endswith(']'):
                if self.__section is not None:
                    self.__end_section()
                self.__section = line[1:-1]

            self.__section_lines.append(line)

        self.__end_section()

        for edit in self.edits:
            if edit['section'] not in self.__sections:
                assert type(edit['section']) is str
                self.__section = edit['section']
                self.__section_lines = ['[%s]' % edit['section']]
                self.__end_section()

    def __end_section(self) -> None:
        assert type(self.__section) is str
        self.__sections.add(self.__section)

        for edit in self.edits:
            if edit['section'] != self.__section:
                continue

            logger.debug('editing %s', self.__section)
            extra_lines = []

            replace_key = edit.get('replace_key', {})
            assert type(replace_key) is dict
            for k, v in sorted(replace_key.items()):
                logger.debug('replacing %s with %s', k, v)
                self.__section_lines = [
                    line for line in self.__section_lines
                    if not line.startswith(k + '=')
                ]
                extra_lines.append('%s=%s' % (k, v))

            for pattern in sorted(edit.get('delete_matched', [])):
                logger.debug('deleting lines matching %s', pattern)
                self.__section_lines = [
                    line for line in self.__section_lines
                    if not fnmatch.fnmatchcase(line, pattern)
                ]

            for pattern in edit.get('comment_out_matched', []):
                logger.debug('commenting out lines matching %s', pattern)
                for i in range(len(self.__section_lines)):
                    if fnmatch.fnmatchcase(self.__section_lines[i], pattern):
                        assert type(edit['comment_out_reason']) is str
                        self.__section_lines[i] = ';' + self.__section_lines[i]
                        self.__section_lines.insert(
                            i, '; ' + edit['comment_out_reason'],
                        )

            for append in edit.get('append_unique', []):
                logger.debug('appending unique line %s', append)
                for line in self.__section_lines:
                    if line == append:
                        break
                else:
                    extra_lines.append(append)

            i = len(self.__section_lines) - 1

            while i >= 0:
                if self.__section_lines[i]:
                    # _s_l[i] is the last non-empty line, insert after it
                    self.__section_lines[i + 1:i + 1] = extra_lines
                    break
                i -= 1
            else:
                # no non-empty lines, insert after the section-opening heading
                self.__section_lines[1:1] = extra_lines

        self.lines.extend(self.__section_lines)
        self.__section_lines = []
        self.__section = None

    def save(self, writer: TextIO) -> None:
        for line in self.lines:
            print(line, file=writer)


class FullLauncher(Launcher):
    def set_id(self) -> None:
        self.keyfile = GLib.KeyFile()
        desktop = os.path.join(RUNTIME_BUILT, self.id + '.desktop')
        if os.path.exists(desktop):
            self.keyfile.load_from_file(desktop, GLib.KeyFileFlags.NONE)
        else:
            self.keyfile.load_from_data_dirs(
                    'applications/%s.desktop' % self.id,
                    GLib.KeyFileFlags.NONE)

        self.name = self.keyfile.get_string(
            GLib.KEY_FILE_DESKTOP_GROUP,
            GLib.KEY_FILE_DESKTOP_KEY_NAME,
        )
        logger.debug('Name: %s', self.name)
        GLib.set_application_name(self.name)

        self.icon_name = self.keyfile.get_string(
            GLib.KEY_FILE_DESKTOP_GROUP,
            GLib.KEY_FILE_DESKTOP_KEY_ICON,
        )
        logger.debug('Icon: %s', self.icon_name)

        try:
            override_id = self.keyfile.get_string(
                GLib.KEY_FILE_DESKTOP_GROUP,
                'X-GameDataPackager-ExpansionFor',
            )
        except GLib.Error:
            pass
        else:
            if self.expansion_name is None:
                self.expansion_name = self.id

            assert type(self.expansion_name) is str
            if self.expansion_name.startswith(override_id + '-'):
                self.expansion_name = self.expansion_name[
                    len(override_id) + 1:
                ]

            self.id = override_id

    def exec_game(self, _unused: Any = None) -> None:
        library_symlinks = self.data.get('library_symlinks', {})
        if 'destination_subdirectory' in library_symlinks:
            dest_subdir = library_symlinks.get('destination_subdirectory')
            os.makedirs(
                os.path.join(self.dot_directory, dest_subdir),
                exist_ok=True,
            )

            arch = self.data.get('arch', 'native')
            libdirs = [
                '/usr/lib/i386-linux-gnu',
                '/usr/lib32',
                '/usr/lib',
            ]
            if (
              (arch == 'native' and os.uname().machine == 'x86_64') or
              arch == 'amd64'
              ):
                libdirs = [
                    '/usr/lib/x86_64-linux-gnu',
                    '/usr/lib64',
                    '/usr/lib',
                ]
            elif (
              (arch == 'native' and os.uname().machine == 'aarch64') or
              arch == 'arm64'
              ):
                libdirs = [
                    '/usr/lib/aarch64-linux-gnu',
                    '/usr/lib64',
                    '/usr/lib',
                ]

            symlinks = library_symlinks.get('symlinks', {})
            for symlink_src, symlink_dst in symlinks.items():
                for d in libdirs:
                    f = os.path.join(d, symlink_src)

                    if os.path.exists(f):
                        dest = os.path.join(
                            self.dot_directory,
                            dest_subdir,
                            symlink_dst,
                        )

                        with suppress(FileNotFoundError):
                            os.remove(dest)

                        os.symlink(f, dest)
                        break

        if self.id == 'ut99':
            if self.expand_tokens.get('UT99System', '') == 'System64':
                dest = os.path.join(self.dot_directory, 'System64')
                with suppress(FileExistsError):
                    os.symlink('System', dest)
            elif self.expand_tokens.get('UT99System', '') == 'SystemARM64':
                dest = os.path.join(self.dot_directory, 'SystemARM64')
                with suppress(FileExistsError):
                    os.symlink('System', dest)

        # Edit before copying, so that we can detect whether this is
        # the first run or not
        for ini, details in self.data.get('edit_unreal_ini', {}).items():
            assert self.dot_directory is not None
            target = os.path.join(self.dot_directory, ini)
            encoding = details.get('encoding', 'windows-1252')

            if os.path.exists(target):
                first_time = False
                try:
                    reader = open(target, encoding='utf-16')
                    reader.readline()
                except UnicodeError:
                    reader = open(target, encoding=encoding)
                else:
                    reader.seek(0)
            else:
                first_time = True

                if os.path.lexists(target):
                    logger.info('Removing dangling symlink %s', target)
                    os.remove(target)

                for base in self.base_directories:
                    if ini in ['System/Rune.ini', 'System/UT2004.ini']:
                        source = os.path.join(base, 'System/Default.ini')
                    else:
                        source = os.path.join(base, ini)

                    if os.path.exists(source):
                        try:
                            reader = open(source, encoding='utf-16')
                            reader.readline()
                        except UnicodeError:
                            reader = open(source, encoding=encoding)
                        else:
                            reader.seek(0)
                        break
                else:
                    raise AssertionError('Required file %s not found', ini)

            if first_time:
                edits = details.get('once', []) + details.get('always', [])
            else:
                edits = details.get('always', [])

            logger.debug('%s', edits)
            editor = IniEditor(edits)

            with reader:
                editor.load(reader)

            d = os.path.dirname(target)

            if d:
                logger.info('Creating directory: %s', d)
                os.makedirs(d, exist_ok=True)

            with open(
                target, 'w', encoding=encoding,
                newline=details.get('newline', '\n')
            ) as writer:
                editor.save(writer)

        super().exec_game()

    def select_language(self) -> None:
        selected_language: str | None = None
        default_language: str | None = None
        start_game: bool = False

        for lang in prefered_langs():
            if lang in self.languages:
                default_language = lang
                break

        def toggle_lang_button(button: Gtk.Widget, data: str) -> None:
            nonlocal selected_language
            selected_language = data

        def click_play_button(button: Gtk.Widget, window: Gtk.Window) -> None:
            nonlocal start_game
            start_game = True
            window.close()

        window = Gtk.Window()
        window.set_default_size(400, 200)
        window.set_title(self.name)
        window.set_icon_name(self.icon_name)

        grid = Gtk.Grid(
            column_spacing=6,
            margin_top=12,
            margin_bottom=12,
            margin_start=12,
            margin_end=12,
        )

        image = Gtk.Image.new_from_icon_name(self.icon_name)
        image.set_pixel_size(128)
        image.set_valign(Gtk.Align.START)
        grid.attach(image, 0, 0, 1, 10)

        subgrid = Gtk.Grid(row_spacing=6)

        langlabel = Gtk.Label(label="Choose your language:")
        langlabel.set_hexpand(True)
        subgrid.attach(langlabel, 0, 0, 1, 1)

        lang = self.languages[0]
        langradio = Gtk.CheckButton(label=getattr(LanguageCodes, lang))
        langradio.connect('toggled', toggle_lang_button, lang)
        if lang == default_language:
            langradio.set_active(True)
        subgrid.attach(langradio, 0, 1, 1, 1)

        i = 2
        for lang in self.languages[1:]:
            radiobutton = Gtk.CheckButton(label=getattr(LanguageCodes, lang))
            radiobutton.set_group(langradio)
            radiobutton.connect('toggled', toggle_lang_button, lang)
            if lang == default_language:
                radiobutton.set_active(True)
            subgrid.attach(radiobutton, 0, i, 1, 1)
            i += 1
        play_button = Gtk.Button.new_with_label('Play')
        play_button.connect('clicked', click_play_button, window)
        subgrid.attach(play_button, 0, i, 1, 1)

        grid.attach(subgrid, 1, 0, 1, 1)

        window.set_child(grid)
        window.present()
        play_button.grab_focus()

        context = GLib.MainContext.default()
        while len(Gtk.Window.get_toplevels()) > 1:
            context.iteration(True)

        if not start_game:
            # Closing the window without clicking on "Play"
            sys.exit(0)

        self.argv.insert(1, self.data.get("languages").get('arg'))
        assert selected_language
        self.argv.insert(2, selected_language)

    def write_confirm_binary_only_stamp(self) -> None:
        assert type(self.warning_stamp) is str
        open(self.warning_stamp, 'a').close()

    def __init__(self) -> None:
        super().__init__()

        # Migrate old directories to new, e.g. ~/.loki/ut -> ~/.utpg
        for old in self.old_dot_directories:
            new = self.dot_directory
            assert new is not None
            os.makedirs(os.path.dirname(new), exist_ok=True)

            have_old = os.path.exists(old)
            have_new = os.path.exists(new)

            if have_old and not have_new:
                logger.debug('Migrating from %s to %s', old, new)
                try:
                    os.rename(old, new)
                except OSError as e:
                    logger.debug(
                        'Unable to rename, falling back to symlink: %s',
                        e,
                    )
                    rel = os.path.relpath(old, start=os.path.dirname(new))
                    logger.debug('Creating symlink %s -> %s', new, rel)
                    os.symlink(rel, new)

        self.window = Gtk.Window()
        self.window.set_default_size(600, 300)
        self.window.set_title(self.name)
        self.window.set_icon_name(self.icon_name)

        self.grid = Gtk.Grid(
            row_spacing=6,
            column_spacing=6,
            margin_top=12,
            margin_bottom=12,
            margin_start=12,
            margin_end=12,
        )
        self.window.set_child(self.grid)

        image = Gtk.Image.new_from_icon_name(self.icon_name)
        image.set_pixel_size(48)
        image.set_valign(Gtk.Align.START)
        self.grid.attach(image, 0, 0, 1, 1)

        self.text_view = Gtk.TextView(
            editable=False,
            cursor_visible=False,
            hexpand=True,
            vexpand=True,
            wrap_mode=Gtk.WrapMode.WORD,
            left_margin=6,
            right_margin=6,
        )
        try:
            self.text_view.set_bottom_margin(6)
            self.text_view.set_top_margin(6)
        except AttributeError:
            logger.warn('You are using an old version of pygobject')
        self.grid.attach(self.text_view, 1, 0, 1, 1)

        subgrid = Gtk.Grid(
            column_spacing=6,
            column_homogeneous=True,
            halign=Gtk.Align.END,
        )

        cancel_button = Gtk.Button.new_with_label('Cancel')
        cancel_button.connect('clicked', lambda _: self.window.close())
        subgrid.attach(cancel_button, 0, 0, 1, 1)

        self.check_box = Gtk.CheckButton.new_with_label("I'll be careful")
        self.check_box.set_hexpand(True)
        self.grid.attach(self.check_box, 0, 1, 2, 1)

        self.ok_button = Gtk.Button.new_with_label('Run')
        self.ok_button.set_sensitive(False)
        subgrid.attach(self.ok_button, 1, 0, 1, 1)

        self.grid.attach(subgrid, 0, 2, 2, 1)

    def run_error(self, message: str) -> None:
        self.show_error(message)

        context = GLib.MainContext.default()
        while len(Gtk.Window.get_toplevels()) > 0:
            context.iteration(True)

    def show_error(self, message: str) -> None:
        self.text_view.get_buffer().set_text(message)
        self.ok_button.set_sensitive(False)
        self.check_box.set_visible(False)
        self.window.present()

    def run_confirm_binary_only(self) -> None:
        self.text_view.get_buffer().set_text(
            self.load_text(
                'confirm-binary-only.txt',
                'Binary-only game, we cannot fix bugs or security '
                'vulnerabilities!'
            )
        )
        self.check_box.bind_property(
            'active', self.ok_button, 'sensitive',
            GObject.BindingFlags.SYNC_CREATE,
        )
        self.ok_button.connect(
            'clicked',
            lambda _: self.__confirm_binary_only_cb(),
        )
        self.window.present()

        context = GLib.MainContext.default()
        while len(Gtk.Window.get_toplevels()) > 0:
            context.iteration(True)

    def __confirm_binary_only_cb(self) -> None:
        try:
            self.write_confirm_binary_only_stamp()
            self.exec_game()
        except Exception:
            self.show_error(traceback.format_exc())


if __name__ == '__main__':
    logging.basicConfig()
    FullLauncher().main()



#!/usr/bin/python3
# encoding=utf-8
#
# Copyright © 2014-2016 Simon McVittie <smcv@debian.org>
# Copyright © 2015-2016 Alexandre Detiste <alexandre@detiste.be>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from abc import (ABCMeta, abstractmethod)
import importlib
import os
import string
from collections.abc import (Iterable, Iterator, Sequence)
from typing import (TypedDict, TYPE_CHECKING, overload)

from ..data import (Package)

if TYPE_CHECKING:
    from typing import (NotRequired)
    from ..game import (GameData)
    from ..data import (PackageRelation)

Compression = bool | str | list[str]


class RecursiveExpansionMap(dict[str, str]):
    def __getitem__(self, k: str) -> str:
        v = super(RecursiveExpansionMap, self).__getitem__(k)
        return string.Template(v).substitute(self)


class PerPackageState:
    def __init__(
        self,
        package: Package,
        per_package_dir: str,
        download: bool,
    ) -> None:
        self.package = package

        # Installed files that will end up in DEBIAN/md5sums for
        # .deb packaging, e.g.
        # { 'usr/share/games/quake3-data/baseq3/pak0.pk3': '1197ca...' }
        self.md5sums: dict[str, str] = {}

        # Component for package, possibly modified: if the license
        # for a freely redistributable game is missing, we demote it from
        # main or non-free to local (i.e. non-distributable).
        self.component: str = package.component

        self.per_package_dir = per_package_dir
        self.destdir = os.path.join(per_package_dir, 'DESTDIR')

        self.lintian_overrides = set(package.lintian_overrides)

        # Flag from CLI, useful for plugins
        self.download: bool = download


class PackagingSystem(metaclass=ABCMeta):
    ASSETS = '$datadir'
    BINDIR = '$prefix/bin'
    DATADIR = '$prefix/share'
    DOCDIR = '$datadir/doc'
    LICENSEDIR = '$datadir/doc'
    PREFIX = '/usr'
    CHECK_CMD: str
    INSTALL_CMD: list[str]

    # Generic tools needed to build packages
    BUILD_DEP: set[str] = set()

    # Exceptions to our normal heuristic for mapping a tool to a package:
    # the executable tool 'unzip' is in the unzip package, etc.
    #
    # Only exceptions need to be listed.
    #
    # 'NotImplemented' means that this dependency is not packaged by
    # the distro.
    PACKAGE_MAP: dict[str, str | None] = {}

    # Exceptions to our normal heuristic for mapping an abstract package name
    # to a package:
    #
    # - the library 'libfoo.so.0' is in a package that Provides libfoo.so.0
    #   (suitable for RPM)
    # - anything else is in the obvious package name
    RENAME_PACKAGES: dict[str, str] = {}

    # Map from Debian architecture name to packaging system's architecture
    # name.
    # We use Debian architectures as our internal representation, as it
    # - has the most architectures supported
    # - differentiates 'any' from 'all'
    # - is the most tested
    # Architectures not found in this map are assumed to match 1:1.
    ARCH_DECODE: dict[str, str] = dict()

    # Map from packaging system's architecture name to Debian
    # architecture name.
    # Architectures not found in this map are assumed to match 1:1.
    ARCH_ENCODE: dict[str, str] = dict()

    def __init__(self, architecture: str | None = None) -> None:
        # Always a dpkg architecture
        self._architecture: str | None = None

        if architecture is not None:
            self._architecture = self.ARCH_ENCODE.get(
                architecture, architecture,
            )

        # Always a set of dpkg architectures
        self._foreign_architectures: set[str] = set()
        # contexts to use when evaluating format- or distro-specific
        # dependencies, in order by preference
        self._contexts: Sequence[str] = ('generic',)

    def derives_from(self, context: str) -> bool:
        return context in self._contexts

    def read_architecture(self) -> str:
        arch = os.uname()[4]
        self._architecture = arch = {
            'armv7l': 'armhf',
            'armhfp': 'armhf',
            'aarch64': 'arm64',
            'i586': 'i386',
            'i686': 'i386',
            'x86_64': 'amd64',
        }.get(arch, arch)
        return arch

    def get_architecture(self, archs: str = '') -> str:
        '''
        Return the dpkg architecture most suitable for archs, which is
        a whitespace-separated string containing dpkg architectures.
        '''

        primary_arch = self._architecture
        if primary_arch is None:
            primary_arch = self.read_architecture()

        if archs:
            # In theory this should deal with wildcards like linux-any,
            # but it's unlikely to be relevant in practice.
            arch_seq = archs.split()

            if primary_arch in arch_seq or 'any' in arch_seq:
                return primary_arch

            for arch in arch_seq:
                if arch in self._foreign_architectures:
                    return arch

        return primary_arch

    def is_installed(self, package: str) -> bool:
        """Return boolean: is a package with the given name installed?"""
        return (self.current_version(package) is not None)

    def is_available(self, package: str) -> bool:
        """Return boolean: is a package with the given name available
        to apt or equivalent?
        """
        try:
            self.available_version(package)
        except Exception:
            return False
        else:
            return True

    @abstractmethod
    def current_version(self, package: str) -> str | None:
        """Return the version number of the given package as a string,
        or None.
        """
        raise NotImplementedError

    @abstractmethod
    def available_version(self, package: str) -> str | None:
        """Return the version number of the given package available in
        apt or equivalent, or raise an exception if unavailable.
        """
        raise NotImplementedError

    @abstractmethod
    def install_packages(
        self,
        packages: Iterable[str],
        method: str | None = None,
        gain_root: str = 'su',
        force: bool = False,
    ) -> None:
        """Install one or more packages (a list of filenames)."""
        raise NotImplementedError

    @overload
    def substitute(
        self,
        template: str | dict[str, str],
        package: str,
        **kwargs: str,
    ) -> str:
        ...

    @overload
    def substitute(
        self,
        template: None,
        package: str,
        **kwargs: str,
    ) -> None:
        ...

    def substitute(
        self,
        template: str | dict[str, str] | None,
        package: str,
        **kwargs: str,
    ) -> str | None:
        if isinstance(template, dict):
            for c in self._contexts:
                if c in template:
                    template = template[c]
                    break
            else:
                return None

        if template is None:
            return template

        if '$' not in template:
            return template

        return string.Template(template).substitute(
                RecursiveExpansionMap(
                    assets=self.ASSETS,
                    bindir=self.BINDIR,
                    datadir=self.DATADIR,
                    docdir=self.DOCDIR,
                    licensedir=self.LICENSEDIR,
                    pkgdocdir=self._get_pkgdocdir(package),
                    pkglicensedir=self._get_pkglicensedir(package),
                    prefix=self.PREFIX,
                    **kwargs))

    def get_libdir(self, package_arch: str) -> str:
        return '/usr/lib'

    def _get_pkgdocdir(self, package: str) -> str:
        return '/'.join((self.DOCDIR, package))

    def _get_pkglicensedir(self, package: str) -> str:
        return '/'.join((self.LICENSEDIR, package))

    def format_relations(
        self,
        relations: Iterable[PackageRelation],
    ) -> Iterator[str]:
        """Yield a native dependency representation for this packaging system
        for each gdp.data.PackagingRelation in relations.
        """
        for pr in relations:
            if pr.contextual:
                for c in self._contexts:
                    if c in pr.contextual:
                        for x in self.format_relations([pr.contextual[c]]):
                            yield x

                        break
            else:
                yield self.format_relation(pr)

    @abstractmethod
    def format_relation(self, pr: PackageRelation) -> str:
        """Return a native dependency representation for this packaging system
        and the given gdp.data.PackagingRelation. It is guaranteed
        that pr.contextual is empty.
        """
        raise NotImplementedError

    def rename_package(self, dependency: str) -> str:
        """Given an abstract package name, return the corresponding
        package name in this packaging system.

        Abstract package names are mostly the same as for Debian,
        except that libraries are represented as libfoo.so.0.
        """
        return self.RENAME_PACKAGES.get(dependency, dependency)

    def package_for_tool(self, tool: str) -> str | None:
        """Given an executable name, return the corresponding
        package name in this packaging system.
        """
        return self.PACKAGE_MAP.get(tool, tool)

    def tool_for_package(self, package: str) -> str:
        """Given a package name, return the corresponding
        main/unique executable in this packaging system.
        """
        for k, v in self.PACKAGE_MAP.items():
            if v == package:
                return k
        return package

    def merge_relations(
        self,
        package: Package,
        rel: str
    ) -> set[str]:
        return set(self.format_relations(package.relations[rel]))

    def generate_description(
        self,
        game: GameData,
        package: Package,
        component: str | None = None
    ) -> tuple[str, list[str]]:

        longname = package.longname or game.longname
        if component is None:
            component = package.component

        if package.short_description is not None:
            short_desc = package.short_description
        elif package.section == 'games':
            short_desc = 'game %s for %s' % (package.data_type, longname)
        else:
            short_desc = longname

        if package.long_description is not None:
            return (short_desc, package.long_description.splitlines())

        long_desc = []
        long_desc.append('This package was built using game-data-packager.')

        if component == 'local':
            long_desc.append(
                'It contains proprietary game data and must not be '
                'redistributed.'
            )
        elif component == 'non-free':
            long_desc.append(
                'It contains proprietary game data that may be redistributed')
            long_desc.append('only under some conditions.')
        else:
            long_desc.append(
                'It contains free game data and may be redistributed.'
            )

        long_desc.append('')

        if package.description:
            for line in package.description.splitlines():
                long_desc.append(line.rstrip())
            long_desc.append('')

        if game.genre:
            long_desc.append(' Genre: ' + game.genre)

        if package.section == 'doc':
            long_desc.append(' Documentation: ' + longname)
        elif package.expansion_for and package.expansion_for in game.packages:
            game_name = (game.packages[package.expansion_for].longname
                         or game.longname)
            if game_name not in long_desc:
                long_desc.append(' Game: ' + game_name)
            if longname != game_name:
                long_desc.append(' Expansion: ' + longname)
        else:
            long_desc.append(' Game: ' + longname)

        copyright = package.copyright or game.copyright
        assert type(copyright) is str
        copyright = copyright.split(' ', 2)[2]
        if copyright not in long_desc:
            long_desc.append(' Published by: ' + copyright)

        engine = self.substitute(
                package.engine or game.engine,
                package.name)

        if engine and package.data_type not in ('music', 'documentation'):
            if long_desc[-1] != '':
                long_desc.append('')

            if '|' in engine:
                virtual = engine.split('|')[-1].strip()
                has_virtual = (virtual.split('-')[-1] == 'engine')
            else:
                has_virtual = False
            engine = engine.split('|')[0].split('(')[0].strip()
            if engine.startswith('gemrb'):
                engine = 'gemrb'
            if has_virtual:
                long_desc.append('Intended for use with some ' + virtual + ',')
                long_desc.append('such as for example: ' + engine)
            else:
                long_desc.append('Intended for use with: ' + engine)

        if package.used_sources:
            if long_desc[-1] != '':
                long_desc.append('')

            long_desc.append('Built from: ' + ', '.join(package.used_sources))

        return (short_desc, long_desc)

    def get_effective_architecture(self, package: Package) -> str:
        '''
        Return an architecture in this packaging system's representation.
        '''
        arch = package.architecture
        if arch != 'all':
            arch = self.get_architecture(arch)
        return self.ARCH_DECODE.get(arch, arch)

    @abstractmethod
    def build_package(
        self,
        per_package_state: PerPackageState,
        game: GameData,
        destination: str,
        compress: Compression = True,
    ) -> str:
        """Build the .deb or equivalent in destination, and return its
        filename.

        per_package_state.per_package_dir may be used as scratch space.
        It already contains a subdirectory named DESTDIR which has been
        populated with the files to be packaged
        (so it contains DESTDIR/usr, etc.)
        """
        raise NotImplementedError

    def available_version_at_least(
        self,
        package: str,
        desired: str,
        *,
        is_installed: bool | None = None
    ) -> bool:
        """Return true if the current or available version of the given
        package is at least a backport of the desired version.
        """
        # Stub implementation: we don't know how this packaging system
        # compares versions, so assume yes it is.
        return True


class NoPackaging(PackagingSystem):
    """
    A stub PackagingSystem used while checking consistency.
    """

    def __init__(self, architecture: str | None = None) -> None:
        super(NoPackaging, self).__init__(architecture=architecture)
        self._contexts = ('generic',)

    def current_version(self, package: str) -> str | None:
        return None

    def available_version(self, package: str) -> str | None:
        return None

    def install_packages(
        self,
        packages: Iterable[str],
        method: str | None = None,
        gain_root: str = 'su',
        force: bool = False,
    ) -> None:
        pass

    def build_package(
        self,
        per_package_state: PerPackageState,
        game: GameData,
        destination: str,
        compress: Compression = True,
    ) -> str:
        raise NotImplementedError

    def format_relation(self, pr: PackageRelation) -> str:
        assert not pr.contextual
        assert not pr.alternatives
        package = pr.package
        assert package is not None, str(pr)

        if pr.version is not None:
            return '%s (%s %s)' % (package, pr.version_operator, pr.version)

        return package


class PackagingTaskArgs(TypedDict):
    packaging: NotRequired[PackagingSystem | None]
    builder_packaging: NotRequired[PackagingSystem | None]


def get_packaging_system(
    format: str,
    distro: str | None = None,
    architecture: str | None = None
) -> PackagingSystem:
    mod = 'game_data_packager.packaging.{}'.format(format)
    packaging_system = importlib.import_module(mod).get_packaging_system(
        distro,
        architecture=architecture,
    )
    assert isinstance(packaging_system, PackagingSystem)
    return packaging_system


def get_native_packaging_system() -> PackagingSystem:
    # lazy import when actually needed
    from ..version import (FORMAT, DISTRO)
    return get_packaging_system(FORMAT, DISTRO)
